import logging
import time
import numpy as np
import pathlib

from .recorder import Recorder
from aegis_sim import variables
from aegis_sim.utilities.funcs import skip

from aegis_sim.parameterization import parametermanager


class ProgressRecorder(Recorder):
    def __init__(self, odir: pathlib.Path):
        self.odir = odir
        self.init_odir()
        self.time_start = time.time()
        self.init_headers()

    def init_headers(self):
        content = ("step", "ETA", "t1M", "runtime", "stg/min", "popsize")
        with open(self.odir / "progress.log", "ab") as f:
            np.savetxt(f, [content], fmt="%-10s", delimiter="| ")

    def write(self, popsize="?"):
        """Record some information about the time and speed of simulation."""

        if skip("LOGGING_RATE"):
            return

        step = variables.steps

        logging.info(
            "%s / %s / N=%s / simname=%s",
            step,
            parametermanager.parameters.STEPS_PER_SIMULATION,
            popsize,
            variables.custom_config_path.stem,
        )

        # Get time estimations
        time_diff = time.time() - self.time_start

        seconds_per_100 = time_diff / step * 100
        eta = (parametermanager.parameters.STEPS_PER_SIMULATION - step) / 100 * seconds_per_100

        steps_per_min = int(step / (time_diff / 60))

        runtime = self.get_dhm(time_diff)
        time_per_1M = self.get_dhm(time_diff / step * 1000000)
        eta = self.get_dhm(eta)

        # Save time estimations
        content = (step, eta, time_per_1M, runtime, steps_per_min, popsize)
        self.write_to_progress_log(content)

    def write_to_progress_log(self, content):
        """

        # OUTPUT SPECIFICATION
        path: /progress.log
        filetype: txt
        category: log
        description: A table documenting the estimated time of simulation completion (ETA), time to run one million steps (t1M), time since simulation start (runtime), number of simulated steps per minute (stg/min) and population size (popsize).
        trait granularity: N/A
        time granularity: N/A
        frequency parameter: LOGGING_RATE
        structure: A str table with custom separator (` | `).
        """
        with open(self.odir / "progress.log", "ab") as f:
            np.savetxt(f, [content], fmt="%-10s", delimiter="| ")

    @staticmethod
    def get_dhm(timediff):
        """Format time in a human-readable format."""
        d = int(timediff / 86400)
        timediff %= 86400
        h = int(timediff / 3600)
        timediff %= 3600
        m = int(timediff / 60)
        return f"{d}`{h:02}:{m:02}"



import logging
import sys


def get_location_of_func(func):
    module_name = func.__module__
    module = sys.modules[module_name]
    module_path = module.__file__
    return f"{module_path}, line {func.__code__.co_firstlineno}"


def log_debug(func):
    def wrapper(*args, **kwargs):
        logging.debug(f"function {func.__name__} ({get_location_of_func(func)})")
        return func(*args, **kwargs)

    return wrapper


def log_info(func):
    def wrapper(*args, **kwargs):
        logging.info(f"Executing function: {func.__name__}.")
        return func(*args, **kwargs)

    return wrapper


def log_error(func):
    def wrapper(*args, **kwargs):
        logging.error(f"Executing function: {func.__name__}.")
        return func(*args, **kwargs)

    return wrapper



import logging


def set_logging(level):
    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(module)s: %(message)s",
        datefmt="%d/%m/%Y %I:%M:%S",
        level=level,  # logging.DEBUG, logging.INFO, ...
    )



import logging
import pandas as pd
import numpy as np

import pathlib

from aegis_sim.dataclasses.population import Population
from .recorder import Recorder
from aegis_sim import variables

from aegis_sim.parameterization import parametermanager
from aegis_sim.utilities.funcs import steps_to_end, skip


class FeatherRecorder(Recorder):
    def __init__(self, odir: pathlib.Path):
        self.odir_genotypes = odir / "snapshots" / "genotypes"
        self.odir_phenotypes = odir / "snapshots" / "phenotypes"
        self.odir_demography = odir / "snapshots" / "demography"
        self.init_dir(self.odir_genotypes)
        self.init_dir(self.odir_phenotypes)
        self.init_dir(self.odir_demography)

    def write(self, population: Population):
        """Record demographic, genetic and phenotypic data from the current population."""

        # If not final snapshots to be taken, and about to skip or the population is extinct, do not write.
        final_snapshots = parametermanager.parameters.SNAPSHOT_FINAL_COUNT > steps_to_end()
        if not final_snapshots and (skip("SNAPSHOT_RATE") or len(population) == 0):
            return

        step = variables.steps

        logging.debug(f"Snapshots recorded at step {step}.")

        self.write_genotypes(step=step, population=population)
        self.write_phenotypes(step=step, population=population)
        self.write_demography(step=step, population=population)

    def write_genotypes(self, step: int, population: Population):
        """

        # OUTPUT SPECIFICATION
        path: /snapshots/genotypes/{step}.feather
        filetype: feather
        category: genotype
        description: A snapshot of complete binary genomes of all individuals at a certain simulation step.
        trait granularity: individual
        time granularity: snapshot
        frequency parameter: SNAPSHOT_RATE
        structure: A bool matrix
        """
        df_gen = pd.DataFrame(np.array(population.genomes.flatten()))
        df_gen.reset_index(drop=True, inplace=True)
        df_gen.columns = [str(c) for c in df_gen.columns]
        df_gen.to_feather(self.odir_genotypes / f"{step}.feather")

    def write_phenotypes(self, step: int, population: Population):
        # TODO add more info to columns and rows
        """

        # OUTPUT SPECIFICATION
        path: /snapshots/phenotypes/{step}.feather
        filetype: feather
        category: phenotype
        description: A snapshot of complete intrinsic phenotypes of all individuals at a certain simulation step.
        trait granularity: individual
        time granularity: snapshot
        frequency parameter: SNAPSHOT_RATE
        structure: A float matrix
        """
        # TODO bugged, wrong header
        df_phe = pd.DataFrame(population.phenotypes.get())
        df_phe.reset_index(drop=True, inplace=True)
        df_phe.columns = [str(c) for c in df_phe.columns]
        df_phe.to_feather(self.odir_phenotypes / f"{step}.feather")

    def write_demography(self, step: int, population: Population):
        """

        # OUTPUT SPECIFICATION
        path: /snapshots/demography/{step}.feather
        filetype: feather
        category: demography
        description: A recording of life history metrics (age, number of births given, step at which born, current size, sex) of all individuals until a certain simulation step.
        trait granularity: individual
        time granularity: snapshot
        frequency parameter: SNAPSHOT_RATE
        structure: A matrix of ints and floats
        """
        dem_attrs = [
            "ages",
            "births",
            "birthdays",
            # "generations",
            "sizes",
            "sexes",
        ]
        demo = {attr: getattr(population, attr) for attr in dem_attrs}
        df_dem = pd.DataFrame(demo, columns=dem_attrs)
        df_dem.reset_index(drop=True, inplace=True)
        df_dem.to_feather(self.odir_demography / f"{step}.feather")



import unittest
import pathlib
from aegis_sim import run


class TestAegisSim(unittest.TestCase):

    def test_run_simulation(self):
        
        path = pathlib.Path(__file__).absolute().parent / "test_sim.yml"

        try:
            run(custom_config_path=path, pickle_path=None, overwrite=True, custom_input_params={})
        except Exception as e:
            self.fail(f"aegis_sim.run raised an exception: {e}")


if __name__ == "__main__":
    unittest.main()



import dash
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output

tab_content = {
    "tab-educational": dash.html.Div(
        [
            dash.html.P(),
            dash.html.Ul(
                [
                    dash.html.P(
                        "AEGIS offers easy-access exploration of topics pertaining to evolution of aging and life history, as well as broader evolutionary topics such as genetic drift, predator-prey dynamics and population dynamics."
                    ),
                    # dash.html.Li(
                    #     "For educators and students, AEGIS is a unique platform for teaching and learning a wide range of concepts in a more intuitive manner, reducing the reliance on advanced mathematical or niche knowledge."
                    # ),
                    # dash.html.Li(
                    #     "Individual-based models, such as those demonstrating Lotka-Volterra predator-prey dynamics, are already established in education, and simulations of genetic drift or epidemiological models are common."
                    # ),
                    # dash.html.Li(
                    #     "This approach focuses on modeling relevant entities and processes, keeping technical and operational details secondary."
                    # ),
                    # dash.html.Li(
                    #     "AEGIS lowers the barrier to entry by offering a web-based interface or simple installation on personal computers, complete with built-in analytics and familiar visualizations."
                    # ),
                ]
            ),
        ]
    ),
    "tab-theoretical": dash.html.Div(
        [
            dash.html.Ul(
                [
                    dash.html.Li(
                        "Address complex and challenging topics that are difficult to model with traditional, analytical methods."
                    ),
                    dash.html.Li(
                        "Explore intricate genetic interactions, fluctuating environmental conditions, and population and evolutionary dynamics."
                    ),
                    dash.html.Li("Test hypotheses and generalize established theories under relaxed assumptions."),
                    dash.html.Li(
                        "Provide new perspectives on the evolution of accelerated aging and longevity, and study customizable genetic architectures of aging, with or without pleiotropy, covering genetic variants as described in mutation accumulation theory, antagonistic pleiotropy theory, and disposable soma theory."
                    ),
                    dash.html.Li(
                        "Generate simulated datasets for testing and refinement of analytical and computational tools for evolutionary biology."
                    ),
                ]
            )
        ]
    ),
    "tab-applied": dash.html.Div(
        [
            dash.html.Ul(
                [
                    dash.html.P(
                        "As a complementary method that is comparatively inexpensive, fast, and does not require animals, AEGIS offers independent study of empirical findings. It can be used before or after running experiments."
                    ),
                    dash.html.P("Before running experiments:", className="fw-bold mb-1"),
                    dash.html.Li(
                        "Assist in formulating hypotheses (models to-be-tested) by helping think through an evolutionary problem and identify relevant factors."
                    ),
                    dash.html.Li(
                        "Explore interactions among relevant evolutionary factors, gauge their relative impact, and test the plausibility of different hypotheses."
                    ),
                    dash.html.Li("Help estimate relevant evolutionary time scales and effect sizes."),
                    dash.html.P("After running experiments:", className="mt-3 fw-bold mb-1"),
                    dash.html.Li(
                        "Study generality and broader relevance of conducted experiments by simulating scenarios with comparable, loosened, or altered assumptions and comparing outcomes with observations."
                    ),
                ]
            )
        ]
    ),
}

# tab_content = {
#     "tab-1": "Content for Students:\n\nAEGIS is an excellent learning tool, offering students an opportunity to simulate life history traits and observe evolutionary processes in a controlled environment. Whether you're studying genetics, ecology, or evolutionary biology, AEGIS provides hands-on experience with real-world scenarios, enhancing theoretical knowledge with practical applications.",
#     "tab-2": "Content for Theorists:\n\nTheorists can leverage AEGIS to test and refine hypotheses about life history evolution. The tool's flexibility allows for the exploration of various evolutionary scenarios, offering insights into the dynamics of selection, mutation, and environmental pressures. AEGIS serves as a bridge between theoretical models and empirical data.",
#     "tab-3": "Content for Empirical Researchers:\n\nFor empirical researchers, AEGIS provides a powerful platform to simulate complex evolutionary processes. It allows for the testing of hypotheses before conducting real-world experiments, saving time and resources. The simulations can help in designing experiments and interpreting experimental data, making AEGIS an indispensable tool in evolutionary biology research.",
# }


# path_here = pathlib.Path(__file__)
# path_md = path_here.parents[3] / "aegis" / "documentation" / "1 who is aegis for.md"

# with open(path_md, "r") as file_:
#     text = markdown.markdown(file_.read())

# Layout
layout = dbc.Card(
    [
        dbc.CardHeader(
            [
                dbc.Tabs(
                    [
                        dbc.Tab(label="Theoretical research", tab_id="tab-theoretical"),
                        dbc.Tab(label="Applied research", tab_id="tab-applied"),
                        dbc.Tab(label="Learning", tab_id="tab-educational"),
                    ],
                    id="card-tabs",
                    active_tab="tab-theoretical",
                    persistence=True,
                ),
            ]
        ),
        dbc.CardBody(dash.html.P(id="card-content", className="card-text")),
    ]
)


# Callback to update card content based on selected tab
@dash.callback(Output("card-content", "children"), Input("card-tabs", "active_tab"))
def update_card_content(active_tab):
    return tab_content[active_tab]



import numpy as np


def leslie_matrix(s, r):
    leslie = np.diag(s, k=-1)
    leslie[0] = r
    leslie[np.isnan(leslie)] = 0
    return leslie


def interval_leslie(interval_life_table, interval_birth_table, index):
    lt = interval_life_table[index]
    s = (1 + lt.pct_change())[1:]
    bt = interval_birth_table[index]
    r = (bt / lt).fillna(0)
    return leslie_matrix(s, r)


def intrinsic_leslie(intrinsic_mortality, intrinsic_fertility, index):
    """
    Hypothetical Leslie matrix that ignores extrinsic mortality.
    """
    m = intrinsic_mortality.iloc[index]
    s = 1 - m
    r = intrinsic_fertility.iloc[index]
    return leslie_matrix(s[:-1], r)


def leslie_breakdown(leslie):
    eigenvalues, eigenvectors = np.linalg.eig(leslie)
    dominant_index = np.argmax(np.abs(eigenvalues))
    dominant_eigenvector = eigenvectors[:, dominant_index]
    return {
        "growth_rate": np.max(np.abs(eigenvalues)),
        "stable_age_structure": dominant_eigenvector / dominant_eigenvector.sum(),
    }



import dash
import dash_bootstrap_components as dbc
from aegis_sim.utilities.container import Container
from aegis_gui.utilities import utilities
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS


def make_select(selected=None):

    paths = utilities.get_sim_paths()

    if not paths:
        return []

    if selected:
        for path in paths:
            if Container(path).name == selected:
                selected_path = path
                break
    else:
        selected_path = paths[0]

    return [
        dash.html.H6("Copy parameters from previous simulations"),
        dash.html.Div(
            children=dbc.InputGroup(
                children=[
                    dbc.InputGroupText("Simulation ID"),
                    dbc.Select(
                        id="sim-config-select",
                        options=[{"label": path.stem, "value": str(path)} for path in paths],
                        # value=["default"] if "default" in dropdown_options else [],
                        value=paths[0].stem,
                        placeholder="Copy simulation",
                        # className="plot-dropdown",
                        # multiple=True,
                        className="me-2",
                    ),
                ],
            ),
            id="sim-config-select-group",
        ),
        dbc.Button(
            [dash.html.I(className="bi bi-arrow-up-square-fill"), "Copy"],
            id="sim-config-copy",
            className="me-1",
            # outline=True,
            color="primary",
        ),
        dbc.Button(
            [dash.html.I(className="bi bi-x-circle-fill"), "Reset"],
            id="reset-run-button",
            className="me-1",
            # outline=True,
            color="danger",
        ),
    ]


@dash.callback(
    dash.Output({"type": "config-input", "index": dash.ALL}, "value", allow_duplicate=True),
    dash.Input("sim-config-copy", "n_clicks"),
    dash.State("sim-config-select", "value"),
    dash.State({"type": "config-input", "index": dash.ALL}, "id"),
    dash.State({"type": "config-input", "index": dash.ALL}, "value"),
    prevent_initial_call=True,
)
def reset_configs(n_clicks, filename, ids, current_values):
    """Why not use DEFAULT_PARAMETERS.get_default_parameters()? To ensure that the order of output values is correct."""

    if n_clicks is None:
        return dash.no_update

    container = utilities.get_container(filename=filename)
    config = container.get_config()

    new_values = []

    for id_, current_value in zip(ids, current_values):
        param_name = id_["index"]
        new_value = config[param_name]
        # new_value = param.default
        if new_value == current_value:
            new_values.append(dash.no_update)
        else:
            new_values.append(new_value)

    return new_values



import logging


class Parameter:
    def __init__(
        self,
        key,
        name,
        domain,
        default,
        info,
        dtype,
        drange,
        info_extended=None,
        inrange=lambda x: True,
        serverrange=lambda x: True,
        serverrange_info="",
        evalrange=None,
        presets={},
        show_in_docs=True,
        show_in_gui=True,
        previous_keys=[],
    ):
        self.key = key
        self.name = name
        self.domain = domain
        self.default = default
        self.info = info
        self.info_extended = info_extended
        self.dtype = dtype
        self.drange = drange
        self.inrange = inrange
        self.serverrange = serverrange
        self.serverrange_info = serverrange_info
        self.evalrange = evalrange
        self.presets = presets
        self.show_in_docs = show_in_docs
        self.show_in_gui = show_in_gui
        self.previous_keys = previous_keys

    def convert(self, value):
        if value is None or value == "":
            return self.default
        elif self.dtype == bool:
            return value in ("True", "true", True)
        else:
            return self.dtype(value)

    def valid(self, value):
        # Not valid if wrong data type
        if not isinstance(value, self.dtype):
            logging.error(f"Value {value} is not of valid type {self.dtype} but of type {type(value)}.")
            return False

        # Not valid if not in range
        if not self.inrange(value):
            return False

        # Valid
        return True

    def get_name(self):
        if self.name:
            return self.name
        name = self.key.replace("_", " ").strip().lower()
        return name

    def validate_dtype(self, value):
        can_be_none = self.default is None
        # given custom value is None which is a valid data type
        if can_be_none and value is None:
            return
        # given custom value is of valid data type
        if isinstance(value, self.dtype):
            return
        # given custom value is int but float is valid
        if self.dtype is float and isinstance(value, int):
            return
        raise TypeError(
            f"You set {self.key} to be '{value}' which is of type {type(value)} but it should be {self.dtype} {'or None' if can_be_none else ''}"
        )

    def validate_inrange(self, value):
        if self.inrange(value):
            return
        raise ValueError(f"{self.key} is set to be '{value}' which is outside of the valid range '{self.drange}'.")

    def validate_serverrange(self, value):
        if self.serverrange(value):
            return
        raise ValueError(
            f"{self.key} is set to be '{value}' which is outside of the valid server range '{self.drange}'."
        )

    def generate_full_evalrange(self):
        if self.evalrange is not None:
            n_datapoints = 10
            from numpy import linspace

            return linspace(self.evalrange[0], self.evalrange[1], n_datapoints)



"""
Functions for dynamic generation of documentation from source code.

"""

import pathlib
import pandas as pd
from aegis_sim.parameterization.default_parameters import DEFAULT_PARAMETERS
from aegis_gui.docs import specifications
from aegis_gui.docs.domains import TEXTS_DOMAIN_MD

here = pathlib.Path(__file__).parent


def get_default_parameter_dictlists():
    cols = [
        "key",
        # "name",
        "default",
        "info",
        "dtype",
        "drange",
        "domain",
    ]
    dictlists = [{col: getattr(p, col) for col in cols} for p in DEFAULT_PARAMETERS.values() if p.show_in_docs]
    return dictlists


parameter_specifications_preamble = """
# Parameter specification

There are three kinds of properties in AEGIS: fixed, parametrizable and variable.
- Fixed properties, such as lack of mating preference, is hard-coded in the current version of AEGIS – it is valid for all simulations and all individuals and cannot be changed.
- Parametrizable properties, such as reproduction mode, is parametrizable – it can be set by a parameter (`REPRODUCTION_MODE`) to various values including `sexual` and `asexual`.
- All other properties, e.g. fertility (biological) or abiotic mortality (environmental), are variable – they differ between individuals and they vary over simulation time.  

The table below provides information about all parameters in AEGIS. Here we provide an explanation of table columns.
- `key` specifies the parameter names as they are given in the GUI and as they are expected in the .yml input configuration files.
- `default` specifies the default value of the parameter. This is the value that will be used if the user does not provide any value.
- `info` specifies a short description of the parameter. Note that to understand some of the parameters it is necessary to understand how they modify the behavior of the submodel they affect.
- `dtype` specifies the data type of the parameter value.
- `drange` specifies the valid range of values that can be used for the parameter.
- `domain` specifies the simulation component which the parameter modifies.

"""


def write_dictlists_as_markdown_table(dict_list, path, preamble):
    # Requires tabulate package
    # dict_list like OUTPUT_SPECIFICATIONS; [{col1: val1, col2, val2, ...}, {col1: val1', col2: val2', ...}, ...]
    df = pd.DataFrame(dict_list)
    md = df.to_markdown()
    with open(path, "w") as file_:
        if preamble:
            file_.writelines(preamble)
        file_.writelines(md)


def dict_to_markdown(data_dict, output_file):
    with open(output_file, "w") as md_file:
        for key, value in data_dict.items():
            # Write the key as a markdown heading
            md_file.write(f"## {key.upper()}\n\n")

            # Write each item in the list as a paragraph
            md_file.write(f"{value}\n\n")  # Add double newline to separate paragraphs


write_dictlists_as_markdown_table(
    get_default_parameter_dictlists(),
    path=here / "default_parameters.md",
    preamble=parameter_specifications_preamble,
)
write_dictlists_as_markdown_table(
    specifications.output_specifications,
    path=here / "output_specifications.md",
    preamble=specifications.output_specifications_preamble,
)
dict_to_markdown(
    TEXTS_DOMAIN_MD,
    output_file=here / "submodel_specifications.md",
)



"""
Abstract away genetic architecture.
"""

import numpy as np

from aegis_sim.submodels.genetics.envdrift import Envdrift
from aegis_sim.submodels.genetics.composite.architecture import CompositeArchitecture
from aegis_sim.submodels.genetics.modifying.architecture import ModifyingArchitecture
from aegis_sim import parameterization

from aegis_sim.dataclasses.phenotypes import Phenotypes


class Architect:
    """Wrapper for a genetic architecture"""

    def __init__(
        self,
        BITS_PER_LOCUS,
        PHENOMAP,
        AGE_LIMIT,
        THRESHOLD,
        ploid,
        ENVDRIFT_RATE,
    ):

        if PHENOMAP:
            assert (
                BITS_PER_LOCUS == 1
            ), f"BITS_PER_LOCUS should be 1 if PHENOMAP is specified but it is set to {BITS_PER_LOCUS}"
            architecture = ModifyingArchitecture(
                ploid=ploid,
                PHENOMAP=PHENOMAP,
                AGE_LIMIT=AGE_LIMIT,
            )

        else:
            architecture = CompositeArchitecture(
                ploid=ploid,
                BITS_PER_LOCUS=BITS_PER_LOCUS,
                AGE_LIMIT=AGE_LIMIT,
                THRESHOLD=THRESHOLD,
            )
        self.architecture = architecture
        self.envdrift = Envdrift(ENVDRIFT_RATE=ENVDRIFT_RATE, genome_shape=self.architecture.get_shape())

    def __call__(self, genomes):
        """Translate genomes into an array of phenotypes probabilities."""

        # Apply the envdrift
        envgenomes = self.envdrift.call(genomes.get_array())
        assert envgenomes.shape == genomes.shape()  # envgenome retains the same shape as genome array

        pheno_array = self.architecture.compute(envgenomes)
        assert len(pheno_array) == len(genomes)  # no individuals are lost during the computation

        return Phenotypes(pheno_array)



import numpy as np

from aegis_sim import constants

from aegis_sim.submodels.genetics.modifying.gpm_decoder import GPM_decoder
from aegis_sim.submodels.genetics.modifying.gpm import GPM


class ModifyingArchitecture:
    """

    GUI
    - when pleiotropy is needed
    - when all bits are 0, the phenotypic values are the ones set from parameters (baseline set in parameters);
    vs composite where it would be 0.
    - ... dev still required
    """

    def __init__(self, ploid, PHENOMAP, AGE_LIMIT):
        self.ploid = ploid
        self.PHENOMAP = PHENOMAP
        self.AGE_LIMIT = AGE_LIMIT

        self.gpm_decoder = GPM_decoder(PHENOMAP)

        self.length = self.gpm_decoder.n

        phenolist = self.gpm_decoder.get_total_phenolist()
        self.phenomap = GPM(
            AGE_LIMIT=AGE_LIMIT,
            phenomatrix=None,
            phenolist=phenolist,
        )

    def get_number_of_phenotypic_values(self):
        return self.AGE_LIMIT * constants.TRAIT_N

    def get_number_of_bits(self):
        return self.length * self.ploid.y

    def get_shape(self):
        return (self.ploid.y, self.length, 1)

    def init_genome_array(self, popsize):
        return np.zeros(shape=(popsize, *self.get_shape()), dtype=np.bool_)

    def init_phenotype_array(self, popsize):
        return np.zeros(shape=(popsize, self.get_number_of_phenotypic_values()))

    def compute(self, genomes):

        if genomes.shape[1] == 1:  # Do not calculate mean if genomes are haploid
            genomes = genomes[:, 0]
        else:
            genomes = self.ploid.diploid_to_haploid(genomes)

        return self.phenomap(
            interpretome=genomes.reshape(len(genomes), -1),
            zeropheno=self.init_phenotype_array(
                popsize=len(genomes),
            ),
        )



import numpy as np
import pickle
import pathlib

from aegis_sim.dataclasses.genomes import Genomes
from aegis_sim.dataclasses.phenotypes import Phenotypes
from aegis_sim import submodels


class Population:
    """Population data

    Contains demographic, genetic and phenotypic data of living individuals.
    """

    attrs = (
        "genomes",
        "ages",
        "births",
        "birthdays",
        "generations",
        "phenotypes",
        "infection",
        "sizes",
        "sexes",
    )

    def __init__(
        self,
        genomes: Genomes,
        ages,
        births,
        birthdays,
        phenotypes: Phenotypes,
        infection,
        sizes,
        sexes,
        generations=None,
    ):
        self.genomes = genomes
        self.ages = ages
        self.births = births
        self.birthdays = birthdays
        self.phenotypes = phenotypes
        self.infection = infection
        self.sizes = sizes
        self.sexes = sexes
        self.generations = generations

        if not (
            len(genomes)
            == len(ages)
            == len(births)
            == len(birthdays)
            == len(phenotypes)
            == len(infection)
            == len(sizes)
            == len(sexes)
            # == len(generations)
        ):
            raise ValueError("Population attributes must have equal length")

    def __len__(self):
        """Return the number of living individuals."""
        return len(self.genomes)

    def __getitem__(self, index):
        """Return a subpopulation."""
        return Population(
            genomes=self.genomes.get(individuals=index),
            ages=self.ages[index],
            births=self.births[index],
            birthdays=self.birthdays[index],
            phenotypes=self.phenotypes.get(individuals=index),
            infection=self.infection[index],
            sizes=self.sizes[index],
            sexes=self.sexes[index],
            generations=self.generations[index] if self.generations is not None else None,
        )

    def __imul__(self, index):
        """Redefine itself as its own subpopulation."""
        for attr in self.attrs:
            if attr == "genomes":
                self.genomes.keep(individuals=index)
            elif attr == "phenotypes":
                self.phenotypes.keep(individuals=index)
            elif attr == "generations":
                self.generations = None
            else:
                setattr(self, attr, getattr(self, attr)[index])
        return self

    def __iadd__(self, population):
        """Merge with another population."""

        for attr in self.attrs:
            if attr == "genomes":
                self.genomes.add(population.genomes)
            elif attr == "phenotypes":
                self.phenotypes.add(population.phenotypes)
            elif attr == "generations":
                self.generations = None
            else:
                val = np.concatenate([getattr(self, attr), getattr(population, attr)])
                setattr(self, attr, val)
        return self

    # def shuffle(self):
    #     order = np.random.arange(len(self))
    #     np.random.shuffle(order)
    #     self *= order

    @staticmethod
    def load_pickle_from(path: pathlib.Path):
        assert path.exists(), f"pickle_path {path} does not exist"
        with open(path, "rb") as file_:
            return pickle.load(file_)

    def save_pickle_to(self, path):
        with open(path, "wb") as file_:
            pickle.dump(self, file_)

    @staticmethod
    def initialize(n):
        genomes = Genomes(submodels.architect.architecture.init_genome_array(n))
        ages = np.zeros(n, dtype=np.int32)
        births = np.zeros(n, dtype=np.int32)
        birthdays = np.zeros(n, dtype=np.int32)
        # generations = np.zeros(n, dtype=np.int32)
        generations = None
        phenotypes = submodels.architect.__call__(genomes)
        infection = np.zeros(n, dtype=np.int32)
        sizes = np.zeros(n, dtype=np.float32)
        sexes = submodels.sexsystem.get_sex(n)
        return Population(
            genomes=genomes,
            ages=ages,
            births=births,
            birthdays=birthdays,
            generations=generations,
            phenotypes=phenotypes,
            infection=infection,
            sizes=sizes,
            sexes=sexes,
        )

    @staticmethod
    def make_eggs(offspring_genomes: Genomes, step, offspring_sexes, parental_generations):
        n = len(offspring_genomes)
        eggs = Population(
            genomes=offspring_genomes,
            ages=np.zeros(n, dtype=np.int32),
            births=np.zeros(n, dtype=np.int32),
            birthdays=np.zeros(n, dtype=np.int32) + step,
            # generations=parental_generations + 1,
            generations=None,
            # phenotypes=submodels.architect.__call__(offspring_genomes),
            phenotypes=np.empty(n),
            infection=np.zeros(n, dtype=np.int32),
            sizes=np.zeros(n, dtype=np.float32),
            sexes=offspring_sexes,
        )
        return eggs



import json
import time
import psutil
import numpy as np
import pathlib
import datetime

from aegis_sim.utilities.get_folder_size import get_folder_size_with_du

from aegis_sim import variables
from .recorder import Recorder


class SummaryRecorder(Recorder):
    """

    Records once.
    """

    def __init__(self, odir: pathlib.Path):
        self.odir = odir
        self.init_odir()
        self.time_start = time.time()
        self.extinct = False

        self.memuse = []
        self.pp = psutil.Process()

    def write_output_summary(self):
        """

        # OUTPUT SPECIFICATION
        path: /output_summary.json
        filetype: json
        category: log
        description: A json dictionary documenting a number of simulation traits at the end of the simulation; e.g. extinction status, random seed, time at start, total runtime, median memory use, total storage use.
        trait granularity: N/A
        time granularity: N/A
        frequency parameter: once
        structure: A json dictionary.
        """
        try:
            storage_use = get_folder_size_with_du(self.odir)
        except:
            storage_use = ""

        summary = {
            "extinct": self.extinct,
            "random_seed": variables.random_seed,
            "time_start": self.time_start,
            "runtime": time.time() - self.time_start,
            "memory_use": self.get_median_memuse(),
            "storage_use": storage_use,
        }
        with open(self.odir / "output_summary.json", "w") as f:
            json.dump(summary, f, indent=4)

    def write_input_summary(self, ticker_pid):
        """

        # OUTPUT SPECIFICATION
        path: /input_summary.json
        filetype: json
        category: genotype
        description: A json dictionary documenting a number of simulation traits at the start of the simulation; e.g. random seed, time at start.
        trait granularity:
        time granularity:
        frequency parameter: once
        structure: A json dictionary.
        """
        summary = {
            "random_seed": variables.random_seed,
            "time_start": self.time_start,
            "pickle_path": str(variables.pickle_path),
            "time_start_formatted": datetime.datetime.fromtimestamp(self.time_start).strftime("%Y-%m-%d %H:%M:%S"),
            "pid": self.pp.pid,
            "ticker_pid": ticker_pid,
        }
        with open(self.odir / "input_summary.json", "w") as f:
            json.dump(summary, f, indent=4)

    def record_memuse(self):
        # TODO refine
        memuse_ = self.pp.memory_info()[0] / float(2**20)
        self.memuse.append(memuse_)
        if len(self.memuse) > 1000:
            self.memuse.pop(0)

    def get_median_memuse(self):
        return np.median(self.memuse)



import logging
import pathlib

from aegis_sim.dataclasses.population import Population
from aegis_sim.bioreactor import Bioreactor
from aegis_sim import variables, submodels, parameterization
from aegis_sim.parameterization import parametermanager
from aegis_sim.recording import recordingmanager


def run(custom_config_path, pickle_path, overwrite, custom_input_params):
    init(custom_config_path, overwrite, pickle_path, custom_input_params)

    population = (
        Population.initialize(n=parametermanager.parameters.CARRYING_CAPACITY)
        if pickle_path is None
        else Population.load_pickle_from(pickle_path)
    )

    bioreactor = Bioreactor(population)

    sim(bioreactor=bioreactor)


def init(custom_config_path, overwrite=False, pickle_path=None, custom_input_params={}):
    """

    When testing aegis, initialize all modules using this function, e.g.

    import aegis_sim
    aegis_sim.init("_.yml")

    And then you can safely import any module.
    """

    custom_config_path = pathlib.Path(custom_config_path)

    parametermanager.init(
        custom_config_path=custom_config_path,
        custom_input_params=custom_input_params,
    )
    variables.init(
        variables,
        custom_config_path=custom_config_path,
        pickle_path=pickle_path,
        RANDOM_SEED=parametermanager.parameters.RANDOM_SEED,
    )
    parameterization.init_traits(parameterization)
    submodels.init(submodels, parametermanager=parametermanager)

    recordingmanager.init(custom_config_path, overwrite)
    recordingmanager.initialize_recorders(TICKER_RATE=parametermanager.parameters.TICKER_RATE)


def sim(bioreactor):
    # presim
    recordingmanager.configrecorder.write_final_config_file(parametermanager.final_config)
    recordingmanager.ticker.start_process()
    ticker_pid = recordingmanager.ticker.pid
    assert ticker_pid is not None
    recordingmanager.summaryrecorder.write_input_summary(ticker_pid=recordingmanager.ticker.pid)
    # TODO hacky solution of decrementing and incrementing steps
    variables.steps -= 1
    recordingmanager.featherrecorder.write(bioreactor.population)
    variables.steps += 1

    # sim
    recordingmanager.phenomaprecorder.write()

    while (variables.steps <= parametermanager.parameters.STEPS_PER_SIMULATION) and not recordingmanager.is_extinct():
        recordingmanager.progressrecorder.write(len(bioreactor.population))
        recordingmanager.simpleprogressrecorder.write()
        bioreactor.run_step()
        variables.steps += 1

    # postsim
    recordingmanager.summaryrecorder.write_output_summary()
    logging.info("Simulation finished.")
    recordingmanager.ticker.stop_process()



"""Contains functions for the computation of relevant population genetic metrics"""

import statistics
import itertools
import logging
import numpy as np

from aegis_sim.parameterization import parametermanager

class PopgenStats:
    def __init__(self):
        self.pop_size_history = []

    def record_pop_size_history(self, genomes):
        """Records population sizes at last 1000 steps"""
        if len(self.pop_size_history) >= 1000:
            del self.pop_size_history[0]
        self.pop_size_history.append(len(genomes))

    def calc(self, input_genomes, mutation_rates):
        """Calculates all popgen metrics

        Set sample_size value to 0 or -1 to not perform any sampling.
        """

        # Infer ploidy from genomes
        self.ploidy = input_genomes.shape[1]

        # Data to process
        self.genomes = self.make_3D(input_genomes)
        self.gsample = self.get_genomes_sample()
        self.nsample = 0 if self.gsample is None else len(self.gsample)  # TODO correct for ploidy?

        # Statistics on population
        self.n = self.get_n()
        self.ne = self.get_ne()
        self.allele_frequencies = self.get_allele_frequencies()
        self.genotype_frequencies = self.get_genotype_frequencies()
        self.mean_h_per_bit = self.get_mean_h_per_bit()
        self.mean_h_per_locus = self.get_mean_h_per_locus()
        self.mean_h = self.get_mean_h()
        self.mean_h_per_bit_expected = self.get_mean_h_per_bit_expected()
        self.mean_h_expected = self.get_mean_h_expected()
        self.mu = self.get_mu(mutation_rates)
        self.theta = self.get_theta()
        self.segregating_sites = self.get_segregating_sites(self.genomes, self.ploidy)
        self.reference_genome = self.get_reference_genome(self.genomes)

        # Statistics on sample
        if self.nsample:
            # TODO gsample is essentially haploid, check if that causes issues
            self.reference_genome_gsample = self.get_reference_genome(self.gsample)
            self.segregating_sites_gsample = self.get_segregating_sites(self.gsample, 1)
            self.theta_w = self.get_theta_w()
            self.theta_pi = self.get_theta_pi()
            self.tajimas_d = self.get_tajimas_d()
            self.sfs = self.get_sfs(self.reference_genome_gsample)  # Uses reference genome calculated from sample
            self.theta_h = self.get_theta_h()
            self.fayandwu_h = self.get_fayandwu_h()
        else:
            # TODO will cause Recorder to fail?
            (
                self.reference_genome_gsample,
                self.segregating_sites_gsample,
                self.theta_w,
                self.theta_pi,
                self.tajimas_d,
                self.sfs,
                self.theta_h,
                self.fayandwu_h,
            ) = [None] * 8

    def emit_simple(self):
        # TODO add headers
        # TODO add headers and indices to all output files; and index and column names

        attrs = [
            "n",
            "ne",
            "mu",
            "segregating_sites",
            "segregating_sites_gsample",
            "theta",
            "theta_w",
            "theta_pi",
            "tajimas_d",
            "theta_h",
            "fayandwu_h",
        ]

        if self.ploidy == 2:
            attrs += ["mean_h", "mean_h_expected"]

        return {attr: getattr(self, attr) for attr in attrs}

    def emit_complex(self):
        attrs = [
            "allele_frequencies",
            "genotype_frequencies",
            "sfs",
            "reference_genome",
            "reference_genome_gsample",
        ]

        if self.ploidy == 2:
            attrs += ["mean_h_per_bit", "mean_h_per_locus", "mean_h_per_bit_expected"]

        return {attr: getattr(self, attr) for attr in attrs}

    ####################
    # HELPER FUNCTIONS #
    ####################

    @staticmethod
    def harmonic(i):
        """Returns the i-th harmonic number"""
        return (1 / np.arange(1, i + 1)).sum()

    @staticmethod
    def harmonic_sq(i):
        """Returns the i-th harmonic square"""
        return (1 / np.arange(1, i + 1) ** 2).sum()

    @staticmethod
    def make_3D(input_genomes):
        """Returns genomes array with merged chromosomes

        Methods of PopgenStats require the genomes to be in the form [individual, locus, bit] where, if individuals are diploid,
        the odd bits belong to a virtual first chromosome, while the even bits belong to a virtual second chromosome.
        If individuals are haploid, all bits belong to one virtual chromosome.

        This is contrast with the genomes arrays in the rest of AEGIS which is in form [individual, chromosome, locus, bit]
        so that all bits from one chromosome are in a separate array than those in the second chromosome.
        If individuals are haploid, the chromosome dimension contains only one element.
        """

        n_individuals, ploidy, n_loci, bits_per_locus = input_genomes.shape

        # TODO report when conversion cannot be executed

        if ploidy == 1:
            return input_genomes[:, 0]

        else:
            genomes = np.empty(
                shape=(n_individuals, n_loci, ploidy * bits_per_locus),
                dtype=np.bool_,
            )

            # Odd bits contain bits from chromosome 0
            genomes[:, :, 0::2] = input_genomes[:, 0]

            # Even bits contain bits from chromosome 1
            genomes[:, :, 1::2] = input_genomes[:, 1]

        return genomes

    @staticmethod
    def make_4D(genomes, ploidy):
        """Returns genomes array with chromosomes split along the second dimension"""

        n_individuals, n_loci, n_bits = genomes.shape
        bits_per_locus = n_bits // ploidy

        # TODO report when conversion cannot be executed

        unstaggered = np.empty(
            shape=(n_individuals, ploidy, n_loci, bits_per_locus),
            dtype=np.bool_,
        )

        if ploidy == 1:
            unstaggered[:, 0] = genomes
        if ploidy == 2:
            # Chromosome 0 contains odd bits from input genomes
            unstaggered[:, 0] = genomes[:, :, 0::2]
            # Chromosome 1 contains even bits from input genomes
            unstaggered[:, 1] = genomes[:, :, 1::2]

        return unstaggered

    def get_genomes_sample(self):
        """Returns a random sample of genomes"""
        if self.ploidy > 1:
            # The chromosomes get aligned
            # 3D (individuals, loci, bits): [1, 2, 1, 2, 1, 2, 1, 2, 3, 4, 3, 4, 3, 4, 3, 4] ->
            #
            # 2D (everything, 2): [[1, 1, 1, 1, 3, 3, 3, 3],
            #                      [2, 2, 2, 2, 4, 4, 4, 4]] ->
            #
            # 3D (chromosomes, loci, bits // 2): [[1, 1, 1, 1],
            #    (individuals * 2, ...)           [2, 2, 2, 2],
            #                                     [3, 3, 3, 3],
            #                                     [4, 4, 4, 4]]
            genomes = (
                self.genomes.reshape(-1, 2).transpose().reshape(self.genomes.shape[0] * 2, self.genomes.shape[1], -1)
            )
        else:
            genomes = self.genomes

        # TODO check if change in ploidy has implications for popgen stats

        # Check if there are enough genomes to sample
        if len(genomes) < 2:  # NOTE tajimas_d requires minimum 2
            return None

        # Sample genomes
        if 0 < parametermanager.parameters.POPGENSTATS_SAMPLE_SIZE <= genomes.shape[0]:
            indices = np.random.choice(
                range(genomes.shape[0]),
                parametermanager.parameters.POPGENSTATS_SAMPLE_SIZE,
                replace=False,
            )
            return genomes[indices]
        else:
            return genomes

    ####################################
    # OUTPUT: Statistics on population #
    ####################################

    def get_n(self):
        """Returns the census population size N"""
        return self.pop_size_history[-1]

    def get_ne(self):
        """Returns the effective population size Ne"""
        return statistics.harmonic_mean(self.pop_size_history)

    def get_allele_frequencies(self):
        """Returns the frequency of the 1-allele at every position of the genome"""
        # Aligns the genomes of the population, disregarding chromosomes, and takes the mean
        return self.genomes.reshape(self.genomes.shape[0], -1).mean(0)

    def get_genotype_frequencies(self):
        """Output: [loc1_bit1_freq00, loc1_bit1_freq01, loc1_bit1_freq11, loc1_bit2_freq00, ...]"""
        if self.ploidy == 1:
            return self.allele_frequencies

        len_pop = self.genomes.shape[0]

        # Genotype = Sum of alleles at a position -> [0, 1, 2]
        genotypes_raw = self.genomes.reshape(-1, 2).sum(1).reshape(len_pop, -1).transpose()

        # Counts the frequencies of 0, 1 and 2 across the population
        genotype_freqs = np.array([np.bincount(x, minlength=3) for x in genotypes_raw]).reshape(-1) / len_pop

        return genotype_freqs

    def get_mean_h_per_bit(self):
        """Returns the mean heterozygosity per bit.
        Output: [Hloc1_bit1, Hloc1_bit2, ...] Entries: (bits_per_locus // 2) * nloci"""
        if self.ploidy == 1:
            return None

        return self.genotype_frequencies[1::3]

    def get_mean_h_per_locus(self):
        """Returns the mean heterozygosity per locus.
        Output: [Hloc1, Hloc2, ...] Entries: nloci"""
        if self.ploidy == 1:
            return None

        h_per_bit = self.mean_h_per_bit
        return h_per_bit.reshape(-1, self.genomes.shape[2] // 2).mean(1)

    def get_mean_h(self):
        """Returns the mean heterozygosity of a population.
        Output: H"""
        if self.ploidy == 1:
            return None

        return self.mean_h_per_bit.mean()

    def get_mean_h_per_bit_expected(self):
        """Returns the expected mean heterozygosity per bit under Hardy-Weinberg-Equilibrium.
        Output: [Heloc1_bit1, Heloc1_bit2, ...] Entries: (bits_per_locus // 2) * nloci"""
        if self.ploidy == 1:
            return None

        genotype_freqs_sqrd = self.genotype_frequencies**2
        sum_each_locus = genotype_freqs_sqrd.reshape(-1, 3).sum(1)
        return 1 - sum_each_locus

    def get_mean_h_expected(self):
        """Returns the expected mean heterozygosity per bit under Hardy-Weinberg-Equilibrium.
        Output: He"""
        if self.ploidy == 1:
            return None

        return self.mean_h_per_bit_expected.mean()

    def get_mu(self, mutation_rates):
        """Return the mutation rate µ per gene per generation -> AEGIS-'Locus' interpreted as a gene"""
        return np.mean(mutation_rates)

    def get_theta(self):
        """Returns the adjusted mutation rate theta = 4 * Ne * µ,
        where µ is the mutation rate per gene per generation and Ne is the effective population size"""
        return (self.ploidy * 2) * self.ne * self.mu

    ##############################################
    # OUTPUT: Statistics on population or sample #
    ##############################################

    def get_reference_genome(self, genomes):
        """Returns the reference genome based on which allele is most common at each position.
        Equal fractions -> 0"""
        return np.round(genomes.reshape(genomes.shape[0], -1).mean(0)).astype("int32")

    def get_segregating_sites(self, genomes, ploidy):
        """Returns the number of segregating sites

        Genomes are first aligned and summed at each site across the population.
        A site is segregating if its sum is not equal to either 0 or the population size N.
        """

        if ploidy == 1:
            pre_segr_sites = genomes.reshape(genomes.shape[0], -1).sum(0)
            segr_sites = (
                genomes.shape[1] * genomes.shape[2]
                - (pre_segr_sites == genomes.shape[0]).sum()
                - (pre_segr_sites == 0).sum()
            )

            return segr_sites

        pre_segr_sites = genomes.reshape(-1, 2).transpose().reshape(genomes.shape[0] * 2, -1).sum(0)
        segr_sites = (
            ((genomes.shape[1] * genomes.shape[2]) // 2)
            - (pre_segr_sites == genomes.shape[0] * 2).sum()
            - (pre_segr_sites == 0).sum()
        )
        return segr_sites

    ################################
    # OUTPUT: Statistics on sample #
    ################################

    def get_theta_w(self):
        """Returns Watterson's estimator theta_w"""
        return self.segregating_sites_gsample / self.harmonic(self.nsample - 1)

    def get_theta_pi(self):
        """Returns the estimator theta_pi (based on pairwise differences)"""
        combs = itertools.combinations(range(self.nsample), 2)

        # Aligns chromosomes and count pairwise differences
        genomes_sample_flat = self.gsample.reshape(self.nsample, -1)
        diffs = np.array([(genomes_sample_flat[i[0]] != genomes_sample_flat[i[1]]).sum() for i in combs])
        total_diffs = diffs.sum()
        ncomparisons = diffs.size

        return total_diffs / ncomparisons

    def get_tajimas_d(self):
        """Returns Tajima's D"""

        if self.nsample < 3:
            return None

        pre_d = self.theta_pi - self.theta_w
        segr_sites = self.segregating_sites_gsample

        if segr_sites == 0:
            logging.info("Cannot compute Tajima's D because there are no segregating sites.")
            return

        a_1 = self.harmonic(self.nsample - 1)
        a_2 = self.harmonic_sq(self.nsample - 1)
        b_1 = (self.nsample + 1) / (3 * (self.nsample - 1))
        b_2 = (2 * (self.nsample**2 + self.nsample + 3)) / (9 * self.nsample * (self.nsample - 1))
        c_1 = b_1 - (1 / a_1)
        c_2 = b_2 - ((self.nsample + 2) / (a_1 * self.nsample)) + (a_2 / (a_1**2))
        e_1 = c_1 / a_1
        e_2 = c_2 / ((a_1**2) + a_2)
        d_stdev = ((e_1 * segr_sites) + (e_2 * segr_sites * (segr_sites - 1))) ** 0.5

        return pre_d / d_stdev

    def get_sfs(self, reference_genome):
        """Returns the site frequency spectrum (allele frequency spectrum) of a sample"""
        pre_sfs = self.gsample.reshape(self.nsample, -1).sum(0)
        pre_sfs[np.nonzero(reference_genome)] -= self.nsample
        pre_sfs = np.abs(pre_sfs)
        sfs = np.bincount(pre_sfs, minlength=self.nsample + 1)[:-1]  # TODO what if len(genomes) < sample_size
        return sfs

    def get_theta_h(self):
        """Returns Fay and Wu's estimator theta_h"""
        # sum from i=1 to i=n-1: ( (2 * S_i * i^2) / (n * (n-1)) )
        sfs = self.sfs
        t_h = ((2 * sfs * (np.arange(self.nsample) ** 2)) / (self.nsample * (self.nsample - 1))).sum()
        return t_h

    def get_fayandwu_h(self):
        """Returns Fay and Wu's H"""
        pre_h = self.theta_pi - self.theta_h
        h_stdev = 1  # TODO: Calculate actual variance of h
        return pre_h / h_stdev



"""Infection manager

Infection status:
    0 .. susceptible
    1 .. infected
    -1 .. dead
"""

import math
import numpy as np


class Infection:
    """

    GUI
    Infection is an optional source of mortality.
    [[FATALITY_RATE]] specifies how deadly the infection is; thus if set to 0, no deaths from infection will occur.
    The infection modeling is inspired by the SIR (susceptible-infectious-removed) model.

    Individuals cannot gain immunity, thus can get reinfected many times.
    The probability to die from an infection is constant as long as the individual is infected; there is no incubation period nor disease progression.
    The same is true for recovering from the disease, which is equal to [[RECOVERY_RATE]].

    Both of these are independent of age and genetics.

    The infectious agent can be transmitted from an individual to an individual but can also be contracted from the environment (and can therefore not be fully eradicated).
    The probability to acquire the infection from the environment is equal to [[BACKGROUND_INFECTIVITY]], and from other infected individuals it grows with [[TRANSMISSIBILITY]]
    but also (logistically) with the proportion of the infected population.
    """

    def __init__(self, BACKGROUND_INFECTIVITY, TRANSMISSIBILITY, RECOVERY_RATE, FATALITY_RATE):
        self.BACKGROUND_INFECTIVITY = BACKGROUND_INFECTIVITY
        self.TRANSMISSIBILITY = TRANSMISSIBILITY
        self.RECOVERY_RATE = RECOVERY_RATE
        self.FATALITY_RATE = FATALITY_RATE

    def get_infection_probability(self, infection_density):
        return self.BACKGROUND_INFECTIVITY - 0.5 + 1 / (1 + math.exp(-self.TRANSMISSIBILITY * infection_density))

    def __call__(self, population):
        """
        First try infecting susceptible.
        """

        # TODO Do not change population here, only return new status

        if len(population) == 0:
            return

        probs = np.random.random(len(population))

        # current status
        infected = population.infection == 1
        susceptible = population.infection == 0

        # compute infection probability
        infection_density = infected.sum() / len(population)
        infection_probability = self.get_infection_probability(infection_density=infection_density)

        # recoveries from old infections
        population.infection[infected & (probs < self.RECOVERY_RATE)] = 0

        # fatalities
        # overrides recoveries
        population.infection[infected & (probs < self.FATALITY_RATE)] = -1

        # new infections
        population.infection[susceptible & (probs < infection_probability)] = 1



# BASIC

# TODO add functions that convert mortality to survival to survivorship to counts, etc.


def get_mortality(survival):
    # TODO Ensure that you are slicing the phenotype array at right places
    y = 1 - survival
    return y


def get_survivorship(survival):
    # TODO Ensure that you are slicing the phenotype array at right places
    y = survival.cumprod(axis=1)
    return y


def get_total_mortality(interval_life_table):
    total_survivorship = get_total_survivorship(interval_life_table)
    total_mortality = -total_survivorship.pct_change(axis=1)
    return total_mortality


def get_total_survivorship(interval_life_table):
    # additive age structure records viability of each individual during its lifetime
    # when applied to a discrete cohort, additive age structure have the same shape as the survivorship curve
    # furthermore, if normalized, it is equivalent to the survivorship curve
    # when not applied to a cohort, the same holds if the population is stationary which approximately holds at short time scales
    total_survivorship = interval_life_table.div(interval_life_table.iloc[:, 0], axis=0)  # normalize
    return total_survivorship


# COMPLEX


def get_life_expectancy(phenotypes, AGE_LIMIT):
    # TODO Ensure that you are slicing the phenotype array at right places
    pdf = phenotypes.iloc[:, :AGE_LIMIT]
    survivorship = pdf.cumprod(1)
    y = survivorship.sum(1)
    return y


def get_longevity(survivorship, proportion_alive):
    """
    Longevity (omega) .. age at which proportion_alive of the population is still alive.
    Usually, this is set to 95% or 99%, sometimes called exceptional lifespan.
    """
    return survivorship.ge(proportion_alive).sum(1)



import dash
from aegis_gui.guisettings.GuiSettings import gui_settings
import dash_bootstrap_components as dbc


def make_button(filename):
    return dash.dcc.Link(
        dbc.Button(
            [dash.html.I(className="bi bi-bar-chart-fill"), "Plot"],
            className="me-2",
            color="primary",
        ),
        href=gui_settings.wrap_href(f"plot?sim={filename}"),
    )



from aegis_sim.dataclasses.genomes import Genomes
from aegis_sim.submodels.reproduction.pairing import pairing
from aegis_sim.submodels.reproduction.recombination import recombination


class Reproducer:
    """
    GUI
    Individuals are fertile starting with [[MATURATION_AGE]] (can be 0) until [[REPRODUCTION_ENDPOINT]] (if 0, no REPRODUCTION_ENDPOINT occurs).
    Reproduction can be sexual (with diploid genomes) or asexual (with diploid or haploid genomes).
    When reproduction is sexual, recombination occurs in gametes at a rate of [[RECOMBINATION_RATE]]
    and gametes will inherit mutations at an age-independent rate
    which can be parameterized (genetics-independent) or set to evolve (genetics-dependent).
    Mutations cause the offspring genome bit states to flip from 0-to-1 or 1-to-0.
    The ratio of 0-to-1 and 1-to-0 can be modified using the [[MUTATION_RATIO]].

    If the population is oviparous, [[INCUBATION_PERIOD]] should be set to -1, 1 or greater.
    When it is set to -1, all laid eggs hatch only once all living individuals die.
    When it is set to 0 or greater, eggs hatch after that specified time.
    Thus, when 0, the population has no egg life step.
    """

    # TODO INCUBATION_PERIOD set to -1 or 1 or greater is stupid
    # TODO INCUBATION_PERIOD is not really a part of this submodel, but it is in the documentation. it should be though.

    # TODO probably better to split mutation logic into another domain and cluster together with genetic architecture stuff

    def __init__(self, RECOMBINATION_RATE, REPRODUCTION_MODE, mutator):
        self.RECOMBINATION_RATE = RECOMBINATION_RATE
        self.REPRODUCTION_MODE = REPRODUCTION_MODE
        self.mutator = mutator

    def generate_offspring_genomes(self, genomes, muta_prob, ages, parental_sexes) -> Genomes:

        if self.REPRODUCTION_MODE == "sexual":
            genomes = recombination(genomes, self.RECOMBINATION_RATE)
            # TODO fix muta_prob that is also being split
            genomes, ages, muta_prob = pairing(Genomes(genomes), parental_sexes, ages, muta_prob)

        # TODO muta_prob does not work currently
        genomes = self.mutator._mutate(genomes, muta_prob, ages)
        genomes = Genomes(genomes)
        return genomes



from aegis_sim.utilities.analysis.survival import get_survivorship


def get_fertility_potential(interval_phenotypes, AGE_LIMIT):
    # TODO Ensure that you are slicing the phenotype array at right places
    # TODO Ensure that fertility is 0 before maturity
    fertility = interval_phenotypes.iloc[:, AGE_LIMIT:]
    y = fertility
    return y


def get_fertility(interval_phenotypes, AGE_LIMIT, MATURATION_AGE):
    fertility_potential = get_fertility_potential(interval_phenotypes, AGE_LIMIT)
    fertility_potential = fertility_potential.T.reset_index(drop=True).T
    fertility_potential.loc[:, range(0, MATURATION_AGE)] = 0
    y = fertility_potential
    return y


def get_cumulative_reproduction(interval_phenotypes, AGE_LIMIT, MATURATION_AGE):
    # https://besjournals.onlinelibrary.wiley.com/doi/full/10.1111/2041-210X.13289
    # BUG fertility is 0 before maturity
    survivorship = get_survivorship(interval_phenotypes).T.reset_index(drop=True).T
    fertility = get_fertility(interval_phenotypes, MATURATION_AGE).T.reset_index(drop=True).T
    y = (survivorship * fertility).cumsum(axis=1)
    return y


def get_lifetime_reproduction(interval_phenotypes, AGE_LIMIT, MATURATION_AGE):
    # BUG fertility is 0 before maturity
    survivorship = get_survivorship(interval_phenotypes, AGE_LIMIT).T.reset_index(drop=True).T
    fertility = get_fertility(interval_phenotypes, AGE_LIMIT, MATURATION_AGE).T.reset_index(drop=True).T
    # y = np.sum((np.array(survivorship) * np.array(fertility)), axis=1)
    y = (survivorship * fertility).sum(axis=1)
    return y



import dash
import datetime
from dash import html
import dash_bootstrap_components as dbc
from aegis_sim.utilities.container import Container
from aegis_gui.utilities import utilities
import pathlib
from aegis_gui.pages.tab_simlog import delete, terminate, download, zipp, open_folder, goplot
from aegis_sim.utilities.get_folder_size import get_folder_size_with_du
from aegis_gui.guisettings.GuiSettings import gui_settings


def make_select(selected=None):

    paths = utilities.get_sim_paths()

    if not paths:
        return dash.html.P("No simulations to display.")

    if selected:
        for path in paths:
            if Container(path).name == selected:
                selected_path = path
                break
    else:
        selected_path = paths[0]

    return dbc.InputGroup(
        children=[
            dbc.InputGroupText("Simulation ID"),
            dbc.Select(
                id="sim-select",
                options=[{"label": path.stem, "value": str(path)} for path in paths],
                # value=["default"] if "default" in dropdown_options else [],
                value=str(selected_path),
                placeholder="Select simulations to plot...",
                # className="plot-dropdown",
                # multiple=True,
            ),
        ],
        style={"marginBottom": "1rem"},
    )


# Div to display simulation information
def get_info_div():
    return dash.html.Div(children=[], id="sim_info_div")


# Function to get simulation information
def get_sim_info(path):
    container = Container(path)
    input_summary = container.get_input_summary()
    output_summary = container.get_output_summary()

    time_of_creation = (
        datetime.datetime.fromtimestamp(input_summary["time_start"]).strftime("%Y-%m-%d %H:%M")
        if input_summary
        else None
    )

    time_of_finishing = (
        datetime.datetime.fromtimestamp(output_summary["time_start"]).strftime("%Y-%m-%d %H:%M")
        if output_summary
        else "(no terminal log detected)"
    )

    return {
        # "Path": str(container.basepath),
        "Running": not container.has_ticker_stopped(),
        "Started": time_of_creation,
        "Finished": time_of_finishing,
        "Went extinct": output_summary.get("extinct", "(no terminal log detected)"),
        "Path to prerun simulation": input_summary.get("pickle_path", "(no prerun population used)"),
        "Size": get_folder_size_with_du(path),
    }


# Callback to update sim_info_div
@dash.callback(
    dash.Output("sim_info_div", "children"),
    dash.Input("sim-select", "value"),
)
def update_info_div(selected_path):
    if not selected_path:
        return html.Div("No simulation selected.")

    path = pathlib.Path(selected_path)
    sim_info = get_sim_info(path)

    # Create list items for each item in sim_info
    list_items = []
    for key, value in sim_info.items():
        list_items.append(dbc.ListGroupItem([html.Strong(f"{key}: "), html.Span(str(value))]))

    # list_items.append(dbc.ListGroupItem([get_folder_size_with_du(path)]))
    list_items.append(
        dbc.ListGroupItem(
            [html.Strong("Path: "), html.Span(str(path), id={"type": "config-download-basepath", "index": path.stem})]
        )
    )
    list_items.append(
        dbc.ListGroupItem(
            [
                goplot.make_button(path.stem),
            ],
            style={"display": "flex"},
        )
    )
    list_items.append(
        dbc.ListGroupItem(
            [
                download.make_button(path.stem),
                download.make_dcc(path.stem),
                zipp.get_zip_button_layout(filename=path.stem),
                open_folder.make_button(path) if gui_settings.ENVIRONMENT == "local" else None,
            ],
            style={"display": "flex"},
        )
    )
    if gui_settings.ENVIRONMENT == "local":
        list_items.append(
            dbc.ListGroupItem(
                [
                    delete.make(path.stem),
                    delete.make_modal(),
                    terminate.make_button(filename=path.stem),
                ],
                style={"display": "flex"},
            )
        )

    # Return a dbc.ListGroup with the simulation information
    return dbc.ListGroup(list_items, flush=True)



import dash
from aegis_gui.pages.tab_plot.plot.prep_setup import FIG_SETUP
from aegis_gui.pages.tab_plot.plot.gen_fig import generate_figure
from aegis_gui.pages.tab_plot.plot.prep_fig import make_empty_figure
from aegis_gui.utilities import log


# Define a helper function to handle the logic
def handle_draw_plot(dropdown_values, selected_fig, dark_mode):
    if not dropdown_values or dropdown_values == [None]:
        return dash.no_update, dash.no_update

    drag_maxs = []
    figures = []

    for fig_name in FIG_SETUP:
        if fig_name == selected_fig:  # Only update the figure that matches the selected tab
            figure, max_iloc = generate_figure(fig_name, dropdown_values, iloc=-1, dark_mode=dark_mode)
            figures.append(figure)
            drag_maxs.append(max_iloc)
        else:
            figures.append(dash.no_update)
            drag_maxs.append(dash.no_update)
    return figures, drag_maxs


# TODO take into consideration the existing slider value


@dash.callback(
    dash.Output({"type": "graph-figure", "index": dash.ALL}, "figure"),
    dash.Output({"type": "graph-slider", "index": dash.ALL}, "max"),
    dash.Output({"type": "graph-slider", "index": dash.ALL}, "value"),
    dash.Input("dropdown-multi", "value"),
    dash.Input("figure-select", "value"),
    dash.Input("refresh-figure-data", "n_clicks"),
    dash.State("color-mode-switch", "value"),
    # prevent_initial_call=True,
)
@log.log_debug
def draw_plot(dropdown_values, figure_selected, refresh_figure_data, dark_mode):
    # Return empty figure and no updates for sliders if dropdown_values is None
    if dropdown_values == []:
        # BUG it complains
        n = len(FIG_SETUP)
        empty_figures = [make_empty_figure()] * n  # Return an empty figure for each graph
        return empty_figures, [0] * n, [1] * n

    # Otherwise, handle the plot drawing logic
    figures, drag_maxs = handle_draw_plot(dropdown_values, figure_selected, dark_mode=dark_mode)
    return figures, drag_maxs, drag_maxs



import dash
from dash import html, dash_table
from aegis_gui.docs.specifications import output_specifications
import dash_bootstrap_components as dbc

from aegis_sim.submodels import genetics
import pathlib
import aegis


dash.register_page(__name__, path="/wiki", name="AEGIS | Wiki")


def make_output_specification_table():
    """

    Documentation for plotly datatables: https://dash.plotly.com/datatable
    """
    data = [
        {key: specs.get(key, "!!! nan") for key in output_specifications[0].keys()} for specs in output_specifications
    ]
    columns = [{"id": c, "name": c} for c in output_specifications[0].keys()]
    return dash_table.DataTable(
        data=data,
        columns=columns,
        style_data={"whiteSpace": "normal", "height": "auto"},
    )


def layout() -> html.Div:
    return html.Div(
        [
            html.H1("""Wiki tab"""),
            dbc.Select(
                id="select",
                options=[
                    {"label": "Specification of the genetic architecture in AEGIS", "value": "genarch"},
                    {"label": "Specification of output files", "value": "output"},
                    {"label": "Specification of input parameters", "value": "input"},
                    {"label": "Specification of submodels", "value": "submodels"},
                    # {"label": "Disabled option", "value": "3", "disabled": True},
                ],
                value="input",
                style={"marginBottom": "1rem"},
                persistence=True,
            ),
            html.Div(id="select-container", children=[]),
        ]
    )


@dash.callback(
    dash.Output("select-container", "children"),  # Updates the content of the container div
    [dash.Input("select", "value")],  # Listens to changes in the Select dropdown
)
def update_select_container(selected_value):
    if selected_value == "output":
        return [
            # html.H6("Specification of output files"),
            make_accordion_specs_output_files(),  # Call the appropriate function to return content
        ]
    elif selected_value == "genarch":
        return [
            # html.H6("Specification of the genetic architecture in AEGIS"),
            dash.dcc.Markdown(
                children=get_specification_of_gen_arch(), mathjax=True, dangerously_allow_html=True
            ),  # Display Markdown with MathJax
        ]
    elif selected_value == "input":
        path_to_dynamic = pathlib.Path(aegis.__file__).parent / "documentation" / "dynamic"
        with open(path_to_dynamic / "default_parameters.md", "r") as file_:
            text = file_.read()
        return dash.dcc.Markdown(children=text)
    elif selected_value == "submodels":
        path_to_dynamic = pathlib.Path(aegis.__file__).parent / "documentation" / "dynamic"
        with open(path_to_dynamic / "submodel_specifications.md", "r") as file_:
            text = file_.read()
        return dash.dcc.Markdown(children=text)
    else:
        return html.P("Please select an option.")  # Default message for no selection or invalid option


def make_accordion_item(d):

    return dbc.ListGroup(
        children=[dbc.ListGroupItem(children=[dash.html.Strong(f"{k}: "), v]) for k, v in d.items() if k != "path"],
        flush=True,
    )


def make_accordion_specs_output_files():
    data = [
        {key: specs.get(key, "!!! nan") for key in output_specifications[0].keys()} for specs in output_specifications
    ]
    return dbc.Accordion(
        id="wiki-accordion",
        children=[dbc.AccordionItem(title=d["path"], children=make_accordion_item(d)) for d in data],
        start_collapsed=True,
    )


def get_specification_of_gen_arch():
    path_to_md = pathlib.Path(genetics.__file__).parent / "doc.md"
    with open(path_to_md, "r") as file_:
        return file_.read()



import numpy as np

import pathlib

from aegis_sim.dataclasses.population import Population
from aegis_sim import parameterization

from aegis_sim.utilities.funcs import skip
from .recorder import Recorder
from aegis_sim import submodels


class IntervalRecorder(Recorder):

    def __init__(self, odir: pathlib.Path):
        self.odir = odir / "gui"
        self.init_odir()
        self.init_headers()

    def record(self, population):
        """Record data that is needed by gui."""
        # TODO rename INTERVAL_RATE into something more representative; potentially, restructure the recording rates
        if skip("INTERVAL_RATE") or len(population) == 0:
            return

        self.write_genotypes(population=population)
        self.write_phenotypes(population=population)

    def init_headers(self):
        with open(self.odir / "genotypes.csv", "ab") as f:
            length = submodels.architect.architecture.length
            ploidy = submodels.architect.architecture.ploid.y
            header0 = list(range(length)) * ploidy
            header1 = np.repeat(np.arange(ploidy), length)
            np.savetxt(f, [header0], delimiter=",", fmt="%i")
            np.savetxt(f, [header1], delimiter=",", fmt="%i")

        with open(self.odir / "phenotypes.csv", "ab") as f:
            age_limit = submodels.architect.architecture.AGE_LIMIT
            trait_list = list(parameterization.traits.keys())
            n_traits = len(trait_list)
            header0 = np.repeat(trait_list, age_limit)
            header1 = list(np.arange(age_limit)) * n_traits
            np.savetxt(f, [header0], delimiter=",", fmt="%s")
            np.savetxt(f, [header1], delimiter=",", fmt="%i")

    def write_genotypes(self, population: Population):
        """
        genotypes.csv | Record allele frequency

        # OUTPUT SPECIFICATION
        path: /gui/genotypes.csv
        filetype: csv
        category: genotype
        description: A table of allele frequencies (frequency of 1's in the population for each site) across simulation intervals. Columns are sites, rows are simulation intervals (spanning INTERVAL_RATE steps).
        trait granularity: population mean
        time granularity: snapshot
        frequency parameter: INTERVAL_RATE
        structure: A float matrix.
        """
        with open(self.odir / "genotypes.csv", "ab") as f:
            array = population.genomes.flatten().mean(0)
            np.savetxt(f, [array], delimiter=",", fmt="%1.3e")

    def write_phenotypes(self, population: Population):
        """
        phenotypes.csv | Record median phenotype

        # OUTPUT SPECIFICATION
        path: /gui/genotypes.csv
        filetype: csv
        category: phenotype
        description: A table of median intrinsic phenotypes (median phenotype rate for each trait at each age) across simulation intervals. Columns are traits, rows are simulation intervals (spanning INTERVAL_RATE steps).
        trait granularity: population median
        time granularity: snapshot
        frequency parameter: INTERVAL_RATE
        structure: A float matrix
        """
        with open(self.odir / "phenotypes.csv", "ab") as f:
            array = np.median(population.phenotypes.get(), 0)
            np.savetxt(f, [array], delimiter=",", fmt="%1.3e")



import dash
import yaml

from aegis_sim.utilities.container import Container
from aegis_gui.guisettings.GuiSettings import gui_settings


def read_yml(path):
    with open(path, "r") as f:
        return yaml.safe_load(f)


def get_container(filename):
    return Container(gui_settings.sim_dir / filename)


def get_config_path(filename):
    return gui_settings.sim_dir / f"{filename}.yml"


def get_sim_paths(sim_dir=None, sort=True):
    if sim_dir is None:
        sim_dir = gui_settings.sim_dir
    paths = [p for p in sim_dir.iterdir() if p.is_dir()]
    if sort:
        paths = sorted(paths, key=lambda path: path.name)
    return paths


def get_sims():
    return [p.stem for p in get_sim_paths()]


def sim_exists(filename: str) -> bool:
    paths = get_sim_paths()
    return any(path.stem == filename for path in paths)


def get_icon(icon_name):
    return dash.html.Img(
        src=f"/aegis/assets/icons/{icon_name}.svg",
        width="16",
        height="16",
    )



class Frailty:
    def __init__(self, FRAILTY_MODIFIER, AGE_LIMIT):
        self.FRAILTY_MODIFIER = FRAILTY_MODIFIER
        self.AGE_LIMIT = AGE_LIMIT

    def modify(self, hazard, ages):
        amount = ages / self.AGE_LIMIT
        return hazard * (1 + amount * self.FRAILTY_MODIFIER)



import dash
import dash_bootstrap_components as dbc
from aegis_gui.utilities import utilities
import logging


def make_select(selected=None):
    return dash.html.Div(
        # id="sim-prerun-control",
        children=[
            dash.html.H6(
                [
                    "Start with a prerun simulation",
                    dbc.Switch(
                        # [dash.html.I(className="bi bi-arrow-up-square-fill"), "Use"],
                        id="prerun-enable-button",
                        className="me-1",
                        value=False,
                        # outline=True,
                        # color="primary",
                        style={"display": "inline-block", "marginLeft": "0.8rem"},
                    ),
                ],
            ),
            dash.html.Div(
                children=[
                    dbc.InputGroup(
                        children=[
                            dbc.InputGroupText("Simulation ID"),
                            dbc.Select(
                                id="prerun-sim-select",
                                # options=[{"label": path.stem, "value": str(path)} for path in paths],
                                value=None,
                                # placeholder="Choose an evolved population",
                                className="me-2",
                                disabled=True,
                            ),
                        ],
                    ),
                ],
                style={"marginBottom": "1rem", "display": "inline-block"},
            ),
        ],
    )


@dash.callback(
    dash.Output("prerun-sim-select", "disabled"),
    dash.Output("prerun-sim-select", "value"),
    dash.Output("prerun-sim-select", "options"),
    dash.Input("prerun-enable-button", "value"),
    # dash.State("prerun-sim-select", "options"),
)
def toggle_select(disable_switch):
    """Enable or disable the dbc.Select based on the dbc.Switch state."""
    if dash.ctx.triggered_id is None:
        return dash.no_update, dash.no_update, dash.no_update

    paths = utilities.get_sim_paths()
    # If no simulations available
    if not paths:
        return True, dash.no_update, dash.no_update

    options = [{"label": path.stem, "value": str(path)} for path in paths]

    new_value = options[0]["value"] if disable_switch else None
    logging.debug(f"Toggle select {new_value}.")
    return not disable_switch, new_value, options



from dash import html, dcc, Output, MATCH, Input, State, callback
import yaml
from aegis_gui.utilities import log
from aegis_sim.utilities.container import Container

import dash_bootstrap_components as dbc


def make_button(filename):
    return dbc.Button(
        [html.I(className="bi bi-gear-fill"), "Download configuration"],
        id={"type": "config-download-button", "index": filename},
        value=filename,
        color="secondary",
        className="me-2",
    )


def make_dcc(filename):
    return dcc.Download(id={"type": "config-dcc-download", "index": filename})


@callback(
    Output({"type": "config-dcc-download", "index": MATCH}, "data"),
    Input({"type": "config-download-button", "index": MATCH}, "n_clicks"),
    State({"type": "config-download-basepath", "index": MATCH}, "children"),
)
def config_file_download_button(n_clicks, basepath):
    if n_clicks is None:
        return n_clicks

    container = Container(basepath)
    config = container.get_config()
    return {
        "content": yaml.dump(config),
        "filename": container.name + ".yml",
    }



import pandas as pd
import pathlib
import logging
import json
import yaml
from typing import Union
import numpy as np
import psutil

from aegis_sim.dataclasses.population import Population
from aegis_sim.constants import VALID_CAUSES_OF_DEATH
from aegis_sim.recording.ticker import Ticker


# TODO for analysis:
# TODO add 0 to survivorship, think about edge cases
# TODO clean up indices, columns and dtypes
# TODO be explicit about aggregation function


class Container:
    """
    Reads and reformats output files so they are available for internal and external use (prepare for export).
    """

    def __init__(self, basepath):
        self.basepath = pathlib.Path(
            basepath
        ).absolute()  # If path to config file is /path/_.yml, then basepath is /path/_
        self.name = self.basepath.stem
        self.data = {}
        # self.set_paths()
        self.paths = None
        self.ticker = None

    def set_paths(self):
        # TODO smarter way of listing paths; you are capturing te files with number keys e.g. '6': ... /te/6.csv; that's silly
        # TODO these are repeated elsewhere, e.g. path for ticker
        self.paths = {
            path.stem: path for path in self.basepath.glob("**/*") if path.is_file() and path.suffix == ".csv"
        }
        self.paths["log"] = self.basepath / "progress.log"
        self.paths["ticker"] = self.basepath / "ticker.txt"
        self.paths["simpleprogress"] = self.basepath / "simpleprogress.log"
        self.paths["output_summary"] = self.basepath / "output_summary.json"
        self.paths["input_summary"] = self.basepath / "input_summary.json"
        self.paths["snapshots"] = {}
        for kind in ("demography", "phenotypes", "genotypes"):
            self.paths["snapshots"][kind] = sorted(
                (self.basepath / "snapshots" / kind).glob("*"),
                key=lambda path: int(path.stem),
            )
        self.paths["pickles"] = sorted(
            (self.basepath / "pickles").glob("*"),
            key=lambda path: int(path.stem),
        )
        self.paths["te"] = sorted(
            (self.basepath / "te").glob("*"),
            key=lambda path: int(path.stem),
        )
        self.paths["popsize_before_reproduction"] = self.basepath / "popsize_before_reproduction.csv"
        self.paths["popsize_after_reproduction"] = self.basepath / "popsize_after_reproduction.csv"
        if not self.paths["log"].is_file():
            logging.error(f"No AEGIS log found at path {self.paths['log']}.")

    def get_path(self, name):
        if self.paths is None:
            self.set_paths()
        return self.paths[name]

    def get_record_structure():
        # TODO
        return

    def report(self):
        """Report present and missing files"""
        # TODO
        return

    def export(self):
        """Export all primary data from the container using general formats"""
        # TODO
        return

    @staticmethod
    def stop_process(pid, kind_of_process):
        try:
            logging.info(f"Terminating {kind_of_process} process with PID {pid}...")
            process = psutil.Process(pid)
            process.terminate()  # or process.kill()
            process.wait()  # Optional: Wait for the process to be fully terminated
            logging.info(f"Process with PID {pid} terminated successfully.")
        except psutil.NoSuchProcess:
            logging.warning(f"No process found with PID {pid}.")
        except psutil.AccessDenied:
            logging.warning(f"Access denied when trying to terminate the process with PID {pid}.")
        except Exception as e:
            logging.error(f"An error occurred: {e}")

    def terminate(self):
        pid = self.get_input_summary()["pid"]
        assert pid is not None
        self.stop_process(pid, "simulation")
        tpid = self.get_input_summary()["ticker_pid"]
        assert tpid is not None
        self.stop_process(tpid, "ticker")

    ############
    # METADATA #
    ############

    def get_log(self, reload=True):
        if ("log" not in self.data) or reload:
            df = pd.read_csv(self.get_path("log"), sep="|")
            df.columns = [x.strip() for x in df.columns]

            def dhm_inverse(dhm):
                nums = dhm.replace("`", ":").split(":")
                return int(nums[0]) * 24 * 60 + int(nums[1]) * 60 + int(nums[2])

            # TODO resolve deprecated function
            try:
                df[["ETA", "t1M", "runtime"]].map(dhm_inverse)
            except:
                df[["ETA", "t1M", "runtime"]].applymap(dhm_inverse)
            self.data["log"] = df
        return self.data["log"]

    def get_simple_log(self):
        try:
            with open(self.get_path("simpleprogress"), "r") as file_:
                text = file_.read()
                step, steps_per_simulation = text.split("/")
                return int(step), int(steps_per_simulation)
        except:
            logging.error(f"No simpleprogress.log found at {self.get_path('simpleprogress')}")

    def get_ticker(self):
        if self.ticker is None:
            TICKER_RATE = self.get_final_config()["TICKER_RATE"]
            self.ticker = Ticker(TICKER_RATE=TICKER_RATE, odir=self.get_path("ticker").parent)
        return self.ticker

    def get_config(self):
        if "config" not in self.data:
            path = self.basepath.parent / f"{self.basepath.stem}.yml"
            with open(path, "r") as file_:
                custom_config = yaml.safe_load(file_)
            # default_config = get_default_parameters()
            if custom_config is None:
                custom_config = {}
            # self.data["config"] = {**default_config, **custom_config}
            self.data["config"] = custom_config
        return self.data["config"]

    def get_final_config(self):
        if "final_config" not in self.data:
            path = self.basepath / "final_config.yml"
            with open(path, "r") as file_:
                final_config = yaml.safe_load(file_)
            if final_config is None:
                final_config = {}
            self.data["final_config"] = final_config
        return self.data["final_config"]

    def get_generations_until_interval(self):
        """Return Series of number of generations simulated up until interval i"""
        # TODO beware that snapshots are not timed linearly; there is a bunch of snapshots at the end of the simulation
        aar = self.get_average_age_at_reproduction()
        aar.iloc[0] = np.inf  # No time has passed, so no generations yet
        IR = self.get_final_config()["INTERVAL_RATE"]
        aar = aar.pipe(lambda s: IR / s).cumsum()
        return aar

    def get_output_summary(self) -> Union[dict, None]:
        path = self.get_path("output_summary")
        if path.exists():
            return self._read_json(path)
        return {}

    def get_input_summary(self):
        return self._read_json(self.get_path("input_summary"))

    ##########
    # TABLES #
    ##########

    def get_birth_table_observed_interval(self, normalize=False):
        """
        Observed data.
        Number of births (int) per parental age during an interval of length INTERVAL_RATE.
        columns.name == parental_age (int)
        index.name == interval (int)
        """
        table = self._read_df("age_at_birth")
        if normalize:
            table = table.div(table.sum(1), axis=0)
        table.index.names = ["interval"]
        table.columns.names = ["parental_age"]
        table.columns = table.columns.astype(int)
        return table

    def get_life_table_observed_interval(self, normalize=False):
        """
        Observed data.
        Number of individuals (int) per age class observed during an interval of length INTERVAL_RATE.
        columns.name == age_class (int)
        index.name == interval (int)
        """
        table = self._read_df("additive_age_structure")
        table.index.names = ["interval"]
        table.columns.names = ["age_class"]
        table.columns = table.columns.astype(int)
        # NOTE normalize by sum
        if normalize:
            table = table.div(table.sum(1), axis=0)
        return table

    def get_life_table_observed_snapshot(self, record_index: int, normalize=False):
        """
        Observed data. Series.
        Number of individuals (int) per age class observed at some simulation step captured by the record of index record_index.
        name == count
        index.name == age_class
        """
        AGE_LIMIT = self.get_config()["AGE_LIMIT"]
        table = (
            self.get_demography_observed_snapshot(record_index)
            .ages.value_counts()
            .reindex(range(AGE_LIMIT), fill_value=0)
        )
        table.index.names = ["age_class"]
        return table

    def get_death_table_observed_interval(self, normalize=False):
        """
        Observed data. Has a MultiIndex.
        Number of deaths (int) per age class observed during an interval of length INTERVAL_RATE.
        columns.name == age_class (int)
        index.names == ["interval", "cause_of_death"] (int, str)
        """
        # TODO think about position of axes
        table = (
            pd.concat({causeofdeath: self._read_df(f"age_at_{causeofdeath}") for causeofdeath in VALID_CAUSES_OF_DEATH})
            .swaplevel()
            .sort_index(level=0)
        )
        table.index.names = ["interval", "cause_of_death"]
        table.columns.names = ["age_class"]
        table.columns = table.columns.astype(int)
        return table

    #######################
    # TABLES : derivative #
    #######################

    def get_surv_observed_interval(self):
        # TODO this is not accurate; this assumes that the population is in an equilibrium, or it only works if the life table is sampling across a long period
        lt = self.get_life_table_observed_interval()
        lt = lt.pct_change(axis=1).shift(-1, axis=1).add(1).replace(np.inf, 1)
        return lt

    def get_fert_observed_interval(self):
        lt = self.get_life_table_observed_interval()
        bt = self.get_birth_table_observed_interval()
        return bt / lt

    ##########
    # BASICS #
    ##########

    # TODO add better column and index names

    def get_genotypes_intrinsic_snapshot(self, record_index):
        """
        columns .. bit index
        index .. individual index
        value .. True or False
        """
        # TODO let index denote the step at which the snapshot was taken
        return self._read_snapshot("genotypes", record_index=record_index)

    def get_phenotype_intrinsic_snapshot(self, trait, record_index):
        """
        columns .. phenotypic trait index
        index .. individual index
        value .. phenotypic trait value
        """
        # TODO organize by trait
        # TODO let index denote the step at which the snapshot was taken
        df = self._read_snapshot("phenotypes", record_index=record_index)
        # df.columns = df.columns.str.split("_")
        return df

    def get_demography_observed_snapshot(self, record_index):
        """
        columns .. ages, births, birthdays, generations, sizes, sexes
        index .. individual index
        """
        # TODO let index denote the step at which the snapshot was taken
        return self._read_snapshot("demography", record_index=record_index)

    def get_genotypes_intrinsic_interval(self, reload=True):
        """
        columns .. bit index
        index .. record index
        value .. mean bit value
        """
        # TODO check that they exist
        df = pd.read_csv(self.get_path("genotypes"), header=[0, 1], index_col=None)
        df.index = df.index.astype(int)
        df.columns = df.columns.set_levels([df.columns.levels[0].astype(int), df.columns.levels[1].astype(int)])
        df.index.names = ["interval"]
        df.columns.names = ["bit_index", "ploidy"]
        return df

    def get_phenotype_intrinsic_interval(self, trait, reload=True):
        """
        columns .. age
        index .. record index
        value .. median phenotypic trait value
        """
        # TODO check that they exist
        df = pd.read_csv(self.get_path("phenotypes"), header=[0, 1])
        df.index.names = ["interval"]
        df.index = df.index.astype(int)
        df.columns.names = ["trait", "age_class"]
        # TODO age_class is str
        return df.xs(trait, axis=1)

    def get_survival_analysis_TE_observed_interval(self, record_index):
        """
        columns .. T, E
        index .. individual
        value .. age at event, event (1 .. died, 0 .. alive)
        """
        # TODO error with T and E in the record; they are being appended on top
        assert record_index < len(self.get_path("te")), "Index out of range"
        data = pd.read_csv(self.get_path("te")[record_index], header=0)
        data.index.names = ["individual"]
        return data

    def get_population_size_before_reproduction(self):
        data = pd.read_csv(self.get_path("popsize_before_reproduction"), header=None)
        data.index.names = ["steps"]
        data.columns = ["popsize"]
        return data

    def get_population_size_after_reproduction(self):
        data = pd.read_csv(self.get_path("popsize_after_reproduction"), header=None)
        data.index.names = ["steps"]
        data.columns = ["popsize"]
        return data

    ###############
    # DERIVATIVES #
    ###############

    def get_lifetime_reproduction(self):
        survivorship = self.get_surv_observed_interval().cumprod(1)
        fertility = self.get_fert_observed_interval()
        return (survivorship * fertility).sum(axis=1)

    def get_average_age_at_reproduction(self):
        bt = self.get_birth_table_observed_interval()
        n_offspring = bt.sum(1)
        average_age_at_reproduction = (bt * bt.columns).sum(1) / n_offspring
        return average_age_at_reproduction

    #############
    # UTILITIES #
    #############

    def _file_exists(self, stem):
        if self.paths is None:
            self.set_paths()
        return stem in self.paths

    def has_ticker_stopped(self):
        return self.get_ticker().has_stopped()

    def _read_df(self, stem, reload=True):
        file_read = stem in self.data
        # TODO Read also files that are not .csv

        if not self._file_exists(stem):
            logging.error(f"File {self.get_path(stem)} des not exist.")
        elif (not file_read) or reload:
            self.data[stem] = pd.read_csv(self.get_path(stem), header=0)

        return self.data.get(stem, pd.DataFrame())

    @staticmethod
    def _read_json(path):
        if not path.exists():
            logging.warning(f"'{path}' does not exist.")
            return None
        with open(path, "r") as file_:
            return json.load(file_)

    def _read_snapshot(self, record_type, record_index):
        assert record_type in self.get_path("snapshots"), f"No records of '{record_type}' can be found in snapshots"
        assert record_index < len(self.get_path("snapshots")[record_type]), "Index out of range"
        return pd.read_feather(self.get_path("snapshots")[record_type][record_index])

    def _read_pickle(self, record_index):
        assert record_index < len(self.get_path("pickles")), "Index out of range"
        return Population.load_pickle_from(self.get_path("pickles")[record_index])



from dash import html, dcc, Output, Input, State, MATCH, callback
import pathlib
import zipfile
import io
from aegis_gui.utilities import log
import dash_bootstrap_components as dbc


def get_zip_button_layout(filename):
    return html.Div(
        children=[
            dbc.Button(
                [html.I(className="bi bi-file-earmark-zip-fill"), "Download data"],
                id={"type": "zip-download-button", "index": filename},
                value=filename,
                className="me-2",
                color="secondary",
            ),
            dcc.Download(id={"type": "zip-dcc-download", "index": filename}),
        ],
    )


@callback(
    Output({"type": "zip-dcc-download", "index": MATCH}, "data"),
    Input({"type": "zip-download-button", "index": MATCH}, "n_clicks"),
    State({"type": "config-download-basepath", "index": MATCH}, "children"),
    prevent_initial_call=True,
    # running=[(Output("zip-download-button", "disabled"), True, False)] # currently not supported
)
def generate_zip(n_clicks, basepath):
    if n_clicks is None:
        return
    folder_path = pathlib.Path(basepath)
    zip_buffer = zip_folder(folder_path)
    return dcc.send_bytes(zip_buffer.getvalue(), f"{folder_path.name}.zip")


# Zipping function


def zip_folder(folder_path: pathlib.Path) -> io.BytesIO:
    if not folder_path.is_dir():
        raise ValueError("The provided path is not a directory")

    # Create a BytesIO object to hold the zipped data
    zip_buffer = io.BytesIO()

    # Create a Zip archive in memory
    with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for file in folder_path.rglob("*"):
            zip_file.write(file, file.relative_to(folder_path))

    # Seek to the beginning of the BytesIO object
    zip_buffer.seek(0)
    return zip_buffer



import numpy as np
from aegis_sim.dataclasses.genomes import Genomes
from aegis_sim import submodels


def pairing(genomes: Genomes, parental_sexes, ages, muta_prob):
    """Return assorted chromatids."""

    # Get pairs
    males, females = submodels.matingmanager.pair_up_polygamously(parental_sexes)
    assert len(males) == len(females)
    n_pairs = len(males)

    # Which gamete
    male_genomes = genomes.get(individuals=males)
    which_gamete = (np.random.random(n_pairs) < 0.5).astype(np.int32)
    male_gametes = male_genomes[np.arange(n_pairs), which_gamete]

    female_genomes = genomes.get(individuals=females)
    which_gamete = (np.random.random(n_pairs) < 0.5).astype(np.int32)
    female_gametes = female_genomes[np.arange(n_pairs), which_gamete]

    # Unify gametes
    gshape = genomes.shape()
    children = np.empty(shape=(n_pairs, *gshape[1:]), dtype=np.bool_)
    children[np.arange(n_pairs), 0] = male_gametes
    children[np.arange(n_pairs), 1] = female_gametes

    # TODO fix splitting of ages and muta_prob
    return children, ages[females], muta_prob[females]



import functools
import numpy as np


from aegis_sim.parameterization import parametermanager

def stan_age(age):
    return age / parametermanager.parameters.AGE_LIMIT


class GPM_decoder:
    """Converter of simple user input into a genotype-phenotype map (as a list or a matrix).

    --- Example input ---

    PHENOMAP:
        "AP1, 7":
            - - "surv"
            - "agespec"
            - - 0.1
            - - "repr"
            - "agespec"
            - - 0.1
        "MA1, 13":
            - - "surv"
            - "agespec"
            - - 0.1

    ---
    """

    def __init__(self, config_PHENOMAP):
        self.blocks = {}
        self.n = 0

        for namen, quartets in config_PHENOMAP.items():
            # YML dict cannot take tuple as key, so it is fused as f'{name}, {n}'
            name, n = namen.split(",")
            n = int(n)
            self.__add_genblock(name, n)
            for trait, agefunc, funcparam in quartets:
                self.__add_encoding(name, trait, agefunc, funcparam)

    def get_total_phenolist(self):
        phenolist = []
        for block in self.blocks.values():
            phenolist.extend(block.get_phenolist())
        return phenolist

    def __len__(self):
        return self.n

    # PRIVATE
    def __add_genblock(self, name, n):
        assert name not in self.blocks.keys(), f"Block with name {name} already exists."

        genblock = Genblock(name=name, n=n, position=self.n)
        self.blocks[name] = genblock
        self.n += genblock.n

    def __add_encoding(self, name, trait, agefunc, funcparam):
        self.blocks[name].add_encoding(trait, agefunc, funcparam)


class Genblock:
    def __init__(self, name, n, position):
        self.name = name
        self.n = n

        self.position = position

        self.encodings = []

        self.phenolist = []

    def add_encoding(self, trait, agefunc, funcparam):
        encoding = {
            "trait": trait,
            "agefunc": agefunc,
            "funcparam": funcparam,
        }
        self.encodings.append(encoding)
        self.__decode(encoding=encoding)

    def get_phenolist(self):
        for index, trait, age, magnitude in self.phenolist:
            yield index + self.position, trait, age, magnitude

    def __decode(self, encoding):

        def add_to_phenolist(index, trait, age, magnitude):
            self.phenolist.append([index, trait, age, magnitude])

        for i in range(self.n):
            func = Genblock.__resolve_function(encoding["agefunc"], encoding["funcparam"])
            if encoding["agefunc"] == "agespec":
                age = np.random.integers(0, parametermanager.parameters.AGE_LIMIT)
                magnitude = func(age)
                add_to_phenolist(i, encoding["trait"], age, magnitude)
            else:
                for age in range(parametermanager.parameters.AGE_LIMIT):
                    magnitude = func(age)
                    add_to_phenolist(i, encoding["trait"], age, magnitude)

    @staticmethod
    def __resolve_function(func, params):

        scale = params

        # def const(intercept, age):
        #     return intercept

        # def linear(intercept, slope, age):
        #     return intercept + slope * age

        # def gompertz(baseline, aging_rate, age):
        #     """
        #     Exponential increase in mortality.
        #     When aging_rate = 0, there is no aging.
        #     Baseline is intercept.
        #     """
        #     return [-1, 1][aging_rate > 0] * baseline * np.exp(abs(aging_rate) * age)

        # def makeham(intercept, baseline, aging_rate, age):
        #     return intercept + gompertz(baseline=baseline, aging_rate=aging_rate, age=age)

        # def siler(baseline_infancy, slope_infancy, intercept, baseline, aging_rate, age):
        #     return baseline_infancy * np.exp(-slope_infancy * age) + makeham(
        #         intercept, baseline=baseline, aging_rate=aging_rate, age=age
        #     )

        # def _normal(pair):
        #     mean, std = pair
        #     return np.random.normal(mean, std)

        # def _exponential(param):
        #     if param > 0:
        #         return np.random.exponential(param)
        #     else:
        #         return -np.random.exponential(-param)

        def _beta(a=3, b=3):
            return np.random.beta(a, b) * 2 - 1
            # return np.random.beta(a, b)

        def acc(intercept, slope, acc, age):
            x = stan_age(age)
            y = intercept + slope * x + acc * x**3
            return scale * y

        def exp(exp, age):
            x = stan_age(age)
            sign = [-1, 1][exp > 0]
            y = sign * ((x + 1) ** (abs(exp) * 5) - 1)
            return scale * y

        def hump(amplitude, x0, width, age):
            x = stan_age(age)
            midpoint = 0.5
            y = amplitude * np.exp(-((x - midpoint - x0) ** 2) / (0.02 + (width * 0.2) ** 2))
            return scale * y

        def sigm(slope, shift, age):
            x = stan_age(age)
            k = slope * 50
            midpoint = 0.5
            y = 1 / (1 + np.exp((-1) * k * (x - midpoint - shift / 2)))
            return scale * y - scale / 2

        if func == "flat" or func == "agespec":
            return functools.partial(acc, _beta(), 0, 0)
        elif func == "lin":
            return functools.partial(acc, _beta() / 2, _beta(), 0)
        elif func == "acc":
            return functools.partial(acc, _beta(), 0, _beta())
        elif func == "exp":
            return functools.partial(exp, _beta())
        elif func == "hump":
            return functools.partial(hump, _beta(), _beta(1, 1), _beta())
        elif func == "sigm":
            return functools.partial(sigm, _beta(), _beta(1, 1))
        else:
            raise Exception(f"{func} not allowed.")

        # if func == "const" or func == "agespec":
        #     return functools.partial(const, _exponential(params[0]))
        # elif func == "linear":
        #     assert len(params) == 2, "Two parameters are needed for a linear block."
        #     return functools.partial(linear, _exponential(params[0]), _exponential(params[1]))
        # elif func == "gompertz":
        #     assert len(params) == 2, "Two parameters are needed for a Gompertz block."
        #     return functools.partial(gompertz, _exponential(params[0]), _exponential(params[1]))
        # else:
        #     raise Exception(f"{func} not allowed.")

    @staticmethod
    def __resolve_function_old(func, params):

        def const(intercept, age):
            return intercept

        def linear(intercept, slope, age):
            return intercept + slope * age

        def gompertz(baseline, aging_rate, age):
            """
            Exponential increase in mortality.
            When aging_rate = 0, there is no aging.
            Baseline is intercept.
            """
            return [-1, 1][aging_rate > 0] * baseline * np.exp(abs(aging_rate) * age)

        def makeham(intercept, baseline, aging_rate, age):
            return intercept + gompertz(baseline=baseline, aging_rate=aging_rate, age=age)

        def siler(baseline_infancy, slope_infancy, intercept, baseline, aging_rate, age):
            return baseline_infancy * np.exp(-slope_infancy * age) + makeham(
                intercept, baseline=baseline, aging_rate=aging_rate, age=age
            )

        def _normal(pair):
            mean, std = pair
            return np.random.normal(mean, std)

        def _exponential(param):
            if param > 0:
                return np.random.exponential(param)
            else:
                return -np.random.exponential(-param)

        if func == "const" or func == "agespec":
            return functools.partial(const, _exponential(params[0]))
        elif func == "linear":
            assert len(params) == 2, "Two parameters are needed for a linear block."
            return functools.partial(linear, _exponential(params[0]), _exponential(params[1]))
        elif func == "gompertz":
            assert len(params) == 2, "Two parameters are needed for a Gompertz block."
            return functools.partial(gompertz, _exponential(params[0]), _exponential(params[1]))
        else:
            raise Exception(f"{func} not allowed.")



"""Wrapper for phenotype vectors."""

import numpy as np
import pandas as pd
from aegis_sim import parameterization
from aegis_sim import constants
import pathlib


class Phenotypes:
    def __init__(self, array):

        # TODO perform this check somewhere
        # assert array.shape[1] == constants.TRAIT_N * parameterization.parametermanager.parameters.AGE_LIMIT

        # clip phenotype to [0,1] / Apply lo and hi bound
        # TODO ??? extract slicing
        array = Phenotypes.clip_array_to_01(array)

        self.array = array

    def __len__(self):
        return len(self.array)

    # def to_frame(self, AGE_LIMIT):

    #     df = pd.DataFrame(self.array)

    #     # Edit index
    #     # df.reset_index(drop=True, inplace=True)
    #     # df.index.name = "individual"

    #     # Edit columns
    #     traits = constants.EVOLVABLE_TRAITS

    #     # AGE_LIMIT = parameterization.parametermanager.parameters.AGE_LIMIT
    #     ages = [str(a) for a in range(AGE_LIMIT)]
    #     multi_columns = pd.MultiIndex.from_product([traits, ages], names=["trait", "age"])
    #     df.columns = multi_columns

    #     return df

    def get(self, individuals=slice(None), loci=slice(None)):
        return self.array[individuals, loci]

    def add(self, phenotypes):
        self.array = np.concatenate([self.array, phenotypes.array])

    def keep(self, individuals):
        self.array = self.array[individuals]

    def extract(self, trait_name, ages, part=None):

        # TODO Shift some responsibilities to Phenotypes dataclass

        # Generate index of target individuals

        n_individuals = len(self)

        which_individuals = np.arange(n_individuals)

        if part is not None:
            which_individuals = which_individuals[part]

        # Fetch trait in question
        trait = parameterization.traits[trait_name]

        # Reminder.
        # Traits can be evolvable and age-specific (thus different across individuals and ages)
        # Traits can be evolvable and non age-specific (thus different between individuals but same across ages)
        # Traits can be not evolvable (thus same for all individuals at all ages)

        if not trait.evolvable:
            probs = trait.initpheno
        else:
            which_loci = trait.start
            if trait.agespecific:
                shift_by_age = ages[which_individuals]
                which_loci += shift_by_age
            probs = self.get(which_individuals, which_loci)

        # expand values back into an array with shape of whole population
        final_probs = np.zeros(n_individuals, dtype=np.float32)
        final_probs[which_individuals] += probs

        return final_probs

    @staticmethod
    def clip_array_to_01(array):
        for trait in parameterization.traits.values():
            array[:, trait.slice] = Phenotypes.clip_trait_to_01(array[:, trait.slice], trait.name)
        return array

    @staticmethod
    def clip_trait_to_01(array, traitname):
        lo = parameterization.traits[traitname].lo
        hi = parameterization.traits[traitname].hi
        return lo + array * (hi - lo)

    # # Recording functions
    # def to_feather(self):
    #     for_feather = pd.DataFrame(self.array)
    #     for_feather.columns = for_feather.columns.astype(str)
    #     return for_feather

    # @staticmethod
    # def from_feather(path: pathlib.Path):
    #     from_feather = pd.read_feather(path)
    #     return Phenotypes(array=from_feather)



"""Cyclic extrinsic mortality
"""

import math
import logging


class Abiotic:
    """
    # TODO maybe mention which phenomena could be modeled by which shape in the doctsring

    GUI
    Abiotic mortality is an optional source of mortality, useful for modeling death by periodic environmental phenomena such as water availability and temperature.
    It has no effect when [[ABIOTIC_HAZARD_OFFSET]] and [[ABIOTIC_HAZARD_AMPLITUDE]] are set to 0.
    It is modeled using periodic functions with a period of [[ABIOTIC_HAZARD_PERIOD]], amplitude of [[ABIOTIC_HAZARD_AMPLITUDE]],
    shape of [[ABIOTIC_HAZARD_SHAPE]] and constant background mortality of [[ABIOTIC_HAZARD_OFFSET]] (negative or positive).
    Negative hazard is clipped to zero.
    Available hazard shapes (waveforms) are flat, sinusoidal, square, triangle, sawtooth, ramp (backward sawtooth) and instant (Dirac comb / impulse train).

    """

    def __init__(self, ABIOTIC_HAZARD_SHAPE, ABIOTIC_HAZARD_OFFSET, ABIOTIC_HAZARD_AMPLITUDE, ABIOTIC_HAZARD_PERIOD):

        self.ABIOTIC_HAZARD_SHAPE = ABIOTIC_HAZARD_SHAPE
        self.ABIOTIC_HAZARD_OFFSET = ABIOTIC_HAZARD_OFFSET
        self.ABIOTIC_HAZARD_AMPLITUDE = ABIOTIC_HAZARD_AMPLITUDE
        self.ABIOTIC_HAZARD_PERIOD = ABIOTIC_HAZARD_PERIOD

        self.func = {
            "flat": self._flat,
            "sinusoidal": self._sinusoidal,
            "triangle": self._triangle,
            "square": self._square,
            "sawtooth": self._sawtooth,
            "ramp": self._ramp,
            "instant": self._instant,
        }[self.ABIOTIC_HAZARD_SHAPE]

        if self.ABIOTIC_HAZARD_SHAPE == "flat" and self.ABIOTIC_HAZARD_AMPLITUDE > 0 and self.ABIOTIC_HAZARD_OFFSET > 0:
            logging.info(
                """
                Note that under flat cnf.abiotic hazard, amplitude and offset have the same effects;
                the total cnf.abiotic mortality is simply their sum.
                """
            )

    def __call__(self, step):
        return self.func(step) + self.ABIOTIC_HAZARD_OFFSET

    def _flat(self, step):
        return self.ABIOTIC_HAZARD_AMPLITUDE

    def _sinusoidal(self, step):
        return self.ABIOTIC_HAZARD_AMPLITUDE * math.sin(2 * math.pi * step / self.ABIOTIC_HAZARD_PERIOD)

    def _triangle(self, step):
        return self.ABIOTIC_HAZARD_AMPLITUDE * (
            1 - 4 * abs(round(step / self.ABIOTIC_HAZARD_PERIOD - 0.5) - (step / self.ABIOTIC_HAZARD_PERIOD - 0.5))
        )

    def _square(self, step):
        return self.ABIOTIC_HAZARD_AMPLITUDE * (
            1 if (step % self.ABIOTIC_HAZARD_PERIOD) < (self.ABIOTIC_HAZARD_PERIOD / 2) else -1
        )

    def _sawtooth(self, step):
        return self.ABIOTIC_HAZARD_AMPLITUDE * (
            2 * (step / self.ABIOTIC_HAZARD_PERIOD - math.floor(step / self.ABIOTIC_HAZARD_PERIOD + 0.5))
        )

    def _ramp(self, step):
        return self.ABIOTIC_HAZARD_AMPLITUDE * (step % self.ABIOTIC_HAZARD_PERIOD) / self.ABIOTIC_HAZARD_PERIOD

    def _instant(self, step):
        """Mortality function that every ABIOTIC_HAZARD_PERIOD steps kills ABIOTIC_HAZARD_AMPLITUDE of the total living population; step 0 is unaffected"""
        if step == 0 or step % self.ABIOTIC_HAZARD_PERIOD:
            return 0

        mortality = np.random.beta(a=1, b=3, size=1)
        return mortality * self.ABIOTIC_HAZARD_AMPLITUDE



from aegis_sim.recording.featherrecorder import FeatherRecorder
from aegis_sim.recording.flushrecorder import FlushRecorder
from aegis_sim.recording.phenomaprecorder import PhenomapRecorder
from aegis_sim.recording.picklerecorder import PickleRecorder
from aegis_sim.recording.popgenstatsrecorder import PopgenStatsRecorder
from aegis_sim.recording.progressrecorder import ProgressRecorder
from aegis_sim.recording.summaryrecorder import SummaryRecorder
from aegis_sim.recording.terecorder import TERecorder
from aegis_sim.recording.ticker import Ticker
from aegis_sim.recording.intervalrecorder import IntervalRecorder


def extract_output_specification_from_docstring(method):
    """Extract information about the output file created by the method"""
    docstring = method.__doc__
    texts = docstring.split("# OUTPUT SPECIFICATION")[1:]
    for text in texts:
        parsed = {}
        for pair in text.strip().split("\n"):
            k, v = pair.split(":", maxsplit=1)
            parsed[k.strip()] = v.strip()
        yield parsed


output_specifications_preamble = """
# Output specification

The table below contains information about all files generated by AEGIS. Here we provide an explanation of various table columns.
- `path` gives the location of the stored files. Paths are relative, all files are located in the same folder as the configuration file.
- `filetype` specifies the types of output files. AEGIS uses simple text formats (e.g. .csv, .txt, .json, .yml) and common python filetypes (.pickle and .feather).
- `category` specifies the kind of information stored – information about `genotype`, `phenotype`, `demography`, `population genetics` or `log` (logging information).
- `description` contains short descriptions of the files.
- `trait granularity` specifies the level of trait aggregation: `individual` means that data has not been aggregated across individuals, `popultion` means that the data has been aggregated across individuals, either by computing the count, mean or median of the target trait. Note that when individual-level data is recorded, all individual are included in the recording; not just a sample. In the current version, AEGIS does not keep track of individual IDs, thus individuals are not identifiable across recordings.
- `time granularity` specifies the level of time aggregation: `snapshot` means that no data has been aggregated over multiple simulation steps, while `interval` means that an aggregated metric has been computed across multiple simulation steps. Aggregated information has a smaller storage footprint, allowing it to be stored more frequently. 
- `frequency` specifies the parameter that sets the frequency of recording. Recording of all files is customizable, apart from files that are recorded once (marked by `once`). Higher frequency allows more granular inspection of the simulation but requires more storage space.
- `structure` gives more information about the internal format of the files.

"""

output_specifications = [
    specification
    for method in (
        FeatherRecorder.write_genotypes,
        FeatherRecorder.write_phenotypes,
        FeatherRecorder.write_demography,
        FlushRecorder.write_age_at,
        PhenomapRecorder.write,
        PickleRecorder.write,
        PopgenStatsRecorder.write,
        ProgressRecorder.write_to_progress_log,
        SummaryRecorder.write_input_summary,
        SummaryRecorder.write_output_summary,
        TERecorder.write,
        Ticker.write,
        IntervalRecorder.write_genotypes,
        IntervalRecorder.write_phenotypes,
    )
    for specification in extract_output_specification_from_docstring(
        method=method
    )  # This loop is necessary in case there are multiple output specifications in one method
]



import numpy as np


def get_none(container, **_):
    return


def get_ages(container, **_):
    return np.arange(1, container.get_config()["AGE_LIMIT"] + 1)


def get_steps_multiplied(container, **kwargs):
    y = kwargs["y"]
    return np.arange(1, len(y) + 1) * container.get_config()["INTERVAL_RATE"]


def get_steps_non_multiplied(container, **kwargs):
    y = kwargs["y"]
    return np.arange(1, len(y) + 1)



import time
import pathlib

from .recorder import Recorder
from aegis_sim import variables

from aegis_sim.parameterization import parametermanager

class SimpleProgressRecorder(Recorder):
    def __init__(self, odir: pathlib.Path):
        self.odir = odir
        self.init_odir()
        self.time_start = time.time()

        self.progress_path = self.odir / "simpleprogress.log"
        self.write_one(0)
        # self.init_headers()

    # def init_headers(self):
    #     content = ("step", "ETA", "t1M", "runtime", "stg/min", "popsize")
    #     with open(self.odir / "progress.log", "ab") as f:
    #         np.savetxt(f, [content], fmt="%-10s", delimiter="| ")

    def write(self):
        """Record some information about the time and speed of simulation."""

        last_updated = self.check_when_last_updated(file_path=self.progress_path)
        if last_updated > 1:
            step = variables.steps
            self.write_one(step)

    def write_one(self, step):
        message = f"{step}/{parametermanager.parameters.STEPS_PER_SIMULATION}"
        with open(self.progress_path, "w") as file_:
            file_.write(message)

    @staticmethod
    def check_when_last_updated(file_path: pathlib.Path):
        last_modified_time = file_path.stat().st_mtime
        time_since_modification = time.time() - last_modified_time
        seconds = time_since_modification
        return seconds


        # # Get time estimations
        # time_diff = time.time() - self.time_start

        # seconds_per_100 = time_diff / step * 100
        # eta = (parametermanager.parameters.STEPS_PER_SIMULATION - step) / 100 * seconds_per_100

        # steps_per_min = int(step / (time_diff / 60))

        # runtime = self.get_dhm(time_diff)
        # time_per_1M = self.get_dhm(time_diff / step * 1000000)
        # eta = self.get_dhm(eta)

        # # Save time estimations
        # content = (step, eta, time_per_1M, runtime, steps_per_min, popsize)
        # self.write_to_progress_log(content)

    # def write_to_progress_log(self, content):
    #     """

    #     # OUTPUT SPECIFICATION
    #     path: /progress.log
    #     filetype: txt
    #     category: log
    #     description: A table documenting the estimated time of simulation completion (ETA), time to run one million steps (t1M), time since simulation start (runtime), number of simulated steps per minute (stg/min) and population size (popsize).
    #     trait granularity:
    #  time granularity:
    #  frequency parameter: A str table with custom separator (` | `).
    #     """
    #     with open(self.odir / "progress.log", "ab") as f:
    #         np.savetxt(f, [content], fmt="%-10s", delimiter="| ")

    # @staticmethod
    # def get_dhm(timediff):
    #     """Format time in a human-readable format."""
    #     d = int(timediff / 86400)
    #     timediff %= 86400
    #     h = int(timediff / 3600)
    #     timediff %= 3600
    #     m = int(timediff / 60)
    #     return f"{d}`{h:02}:{m:02}"



import dash
import dash_bootstrap_components as dbc
from aegis_gui.pages.tab_plot.plot import prep_y
from aegis_gui.pages.tab_plot.plot import prep_x


FIG_SETUP = {
    "bit states": {
        "title": "bit states",
        "supports_multi": False,
        "prep_y": prep_y.get_bit_states,
        "prep_x": prep_x.get_steps_multiplied,
        "prep_figure": "make_heatmap_figure",
        "description": dash.dcc.Markdown(
            """
            Bit states for each site at a given step.
            \n
            Population averages. Interval averages.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "simulation step",
            "yaxis_title": "genome site",
        },
    },
    "death table": {
        "title": "death table",
        "supports_multi": False,
        "prep_y": prep_y.get_death_table,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_bar_figure_stacked",
        "description": dash.dcc.Markdown(
            """
            Number of deaths per age class, stratified by cause.
            \n
            Interval averages.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "number of deaths",
        },
    },
    "death table normalized": {
        "title": "death table normalized",
        "supports_multi": False,
        "prep_y": prep_y.get_death_table_normalized,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_bar_figure_stacked",
        "description": dash.dcc.Markdown(
            """
            Number of deaths per age class, stratified by cause.
            \n
            Interval averages.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "number of deaths",
            "yaxis": {"range": [0, 1]},
        },
    },
    "derived allele frequencies": {
        "title": "derived allele frequencies",
        "supports_multi": False,
        "prep_y": prep_y.get_derived_allele_freq,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_hist_figure",
        "description": dash.dcc.Markdown(
            """
            Sites with derived allele frequency of 0 are ignored.
            Ancestral states were the most common states an interval ago.
            \n
            """,
            mathjax=True,
        ),
        "nbinsx": 10,
        # graph
        "figure_layout": {
            "xaxis_title": "derived allele frequency",
            "yaxis_title": "number of genome sites",
            "xaxis": {
                "range": [0, 1],
            },
            # Change log scale of y axis
            # "yaxis": {
            #     "type": "log",
            # },
        },
    },
    "intrinsic mortality": {
        "title": "intrinsic mortality",
        "supports_multi": True,
        "prep_y": prep_y.get_mortality_intrinsic,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            """
            Genetic (individual-specific, heritable) mortality at a given age.
            \n
            Population medians.
            """,
            mathjax=True,
        ),
        # dash.html.Span(
        #     "tooltips", id="tooltip-target-1", style={"textDecoration": "underline", "cursor": "pointer"}
        # ),
        # dbc.Tooltip("This is the first tooltip", target="tooltip-target-1"),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "intrinsic mortality",
        },
    },
    "total mortality": {
        "title": "total mortality",
        "supports_multi": True,
        "prep_y": prep_y.get_mortality_observed,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            """
            Observed (individual-specific, heritable) mortality at a given age (all sources of mortality considered).
            \n
            Population averages. Interval averages.
            \n
            Missing data points are for age classes at which no living individuals have been observed.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "total mortality",
        },
    },
    # BUG goes up and down when the snapshot_rate is very low
    "total survivorship": {
        "title": "total survivorship",
        "supports_multi": True,
        "prep_y": prep_y.get_total_survivorship,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            """
            Observed expected probability to survive to a specific age (when all sources of mortality are considered).
            \n
            Population averages. Interval averages.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "total survivorship",
        },
    },
    "intrinsic survivorship": {
        "title": "intrinsic survivorship",
        "supports_multi": True,
        "prep_y": prep_y.get_intrinsic_survivorship,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            # TODO clarify
            """
            Expected probability to survive to a specific age class only given genetic mortality.
            \n
            Computed using the median intrinsic mortalities.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "intrinsic survivorship",
        },
    },
    "life expectancy": {
        "title": "life expectancy at age 0",
        "supports_multi": True,
        "prep_y": prep_y.get_life_expectancy,
        "prep_x": prep_x.get_steps_multiplied,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            # TODO check this
            """
            Expected lifespan at birth. Plotted over the course of the simulation.
            \n
            Population averages.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "simulation step",
            "yaxis_title": "life expectancy",
        },
    },
    "population size": {
        "title": "population size",
        "supports_multi": True,
        "prep_y": prep_y.get_population_size_after_reproduction,
        "prep_x": prep_x.get_steps_non_multiplied,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            # TODO check this
            """
            .
            \n
            .
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "simulation step",
            "yaxis_title": "life expectancy",
        },
    },
    "fertility": {
        "title": "intrinsic fertility",
        "supports_multi": True,
        "prep_y": prep_y.get_fertility,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_line_figure",
        "description": dash.dcc.Markdown(
            """
            The intrinsic probability to produce a single offspring at each age class.
            \n
            Population medians.
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "intrinsic fertility",
        },
    },
    # "cumulative reproduction": {
    #     "title": "cumulative reproduction",
    # "supports_multi": True,
    # "prep_y": prep_y.get_cumulative_reproduction,
    #     "prep_x": prep_x.get_ages,
    #     "prep_figure": "make_line_figure",
    #     "description": dash.dcc.Markdown(
    #         """
    #         The expected number of offspring produced per individual until a given age class.
    #         \n
    #         Population averages.
    #         """,
    #         # \n$\sum_{x=0}^{a^*}m(x)l(x)$
    #         mathjax=True,
    #     ),
    #     # graph
    #     "figure_layout": {
    #         "xaxis_title": "age class",
    #         "yaxis_title": "cumulative number of offspring",
    #     },
    # },
    # "lifetime reproduction": {
    #     "title": "lifetime reproduction",
    # "supports_multi": True,
    # "prep_y": prep_y.get_lifetime_reproduction,
    #     "prep_x": prep_x.get_steps_multiplied,
    #     "prep_figure": "make_line_figure",
    #     "description": dash.dcc.Markdown(
    #         """
    #         The expected number of offspring produced per individual until death. Plotted over the course of the simulation.
    #         \n
    #         Population averages.
    #         """,
    #         mathjax=True,
    #     ),
    #     # graph
    #     "figure_layout": {
    #         "xaxis_title": "simulation step",
    #         "yaxis_title": "lifetime number of offspring",
    #     },
    # },
    "birth table": {
        "title": "birth table",
        "supports_multi": True,
        "prep_y": prep_y.get_birth_table,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_bar_figure_not_stacked",
        "description": dash.dcc.Markdown(
            """
            The number of newborns produced by parents of a given age class.
            \n
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "number of newborns",
        },
    },
    "life table": {
        "title": "life table",
        "supports_multi": True,
        "prep_y": prep_y.get_life_table,
        "prep_x": prep_x.get_ages,
        "prep_figure": "make_bar_figure_not_stacked",
        "description": dash.dcc.Markdown(
            """
            The proportion of living individuals.
            \n
            """,
            mathjax=True,
        ),
        # graph
        "figure_layout": {
            "xaxis_title": "age class",
            "yaxis_title": "proportion of living individuals",
        },
    },
    # "death structure": {
    #     "title": "death structure",
    # "supports_multi": True,
    # "prep_y": prep_y.get_death_structure,
    #     "prep_x": prep_x.get_ages,
    #     "prep_figure": "make_line_figure",
    #     "description": dash.dcc.Markdown(
    #         """The measured ratio of intrinsic deaths versus total deaths, grouped by age.""",
    #         mathjax=True,
    #     ),
    #     # graph
    #     "figure_layout": {
    #         "xaxis_title": "age class",
    #         "yaxis_title": "",
    #     },
    # },
    # "total survivorship": {
    #     "title": "total survivorship",
    # "supports_multi": True,
    # "description": dash.dcc.Markdown(
    #         """xxx.""",
    #         mathjax=True,
    #     ),
    #     # graph
    #     "figure_layout": {
    #         "xaxis_title": "age class",
    #         "yaxis_title": "",
    #         # "yaxis": {"range": [0, 1]},
    #     },
    # },
}



import numpy as np


class Resources:

    def __init__(self, CARRYING_CAPACITY):
        self.CARRYING_CAPACITY = CARRYING_CAPACITY
        self.capacity = 0

    def replenish(self):
        self.capacity = self.CARRYING_CAPACITY

    def reduce(self, amount):
        if amount > self.capacity:
            self.capacity = 0
        else:
            self.capacity -= amount

    def scavenge(self, demands):
        """Return the available amount of resources and reduce exploited amount"""

        total_demand = demands.sum()

        if total_demand > self.capacity:
            ration = self.capacity / len(demands)
            self.reduce(total_demand)
            return ration * np.ones(shape=demands.shape)
        else:
            self.reduce(total_demand)
            return demands



from .recorder import Recorder
from aegis_sim.dataclasses.population import Population


class PopsizeRecorder(Recorder):
    def __init__(self, odir):
        self.odir = odir

    def write(self, population: Population, filename: str):
        popsize = len(population)
        path = self.odir / filename
        with open(path, "a") as file_:
            file_.write(f"{popsize}\n")

    def write_before_reproduction(self, population):
        self.write(population, "popsize_before_reproduction.csv")

    def write_after_reproduction(self, population):
        self.write(population, "popsize_after_reproduction.csv")



# test_app.py
import unittest
from aegis_gui import app as app_
from aegis_gui.guisettings.GuiSettings import gui_settings


class TestDashApp(unittest.TestCase):
    def setUp(self):

        gui_settings.set(environment="local", debug=False)

        app = app_.get_app()

        # Set up the test client
        self.client = app.server.test_client()
        self.app_context = app.server.app_context()
        self.app_context.push()
        print(self.app_context)

    def tearDown(self):
        # Clean up after tests
        self.app_context.pop()

    def test_app_runs(self):
        # Send a request to the root URL
        response = self.client.get("/aegis/")

        # Check that the response status code is 200 (OK)
        self.assertEqual(response.status_code, 200)

        # Optionally, check that the response contains expected content
        self.assertIn(b"AEGIS", response.data)


if __name__ == "__main__":
    unittest.main()



import dash
import dash_bootstrap_components as dbc
from aegis_gui import layout
from aegis_gui.guisettings.GuiSettings import gui_settings


@dash.callback(
    [dash.Output(f"link-nav-{page}", "active") for page in ["home", "config", "plot", "simlog", "wiki"]],
    [dash.Input("url", "pathname")],
)
def toggle_active_links(pathname):
    if pathname is None:
        # Default to home if pathname is None
        pathname = "/"
    return [
        pathname == f"/{page}" or (page == "home" and pathname == "/")
        for page in ["home", "config", "plot", "simlog", "wiki"]
    ]


def get_app():
    app = dash.Dash(
        __name__,
        suppress_callback_exceptions=True,
        update_title="",
        url_base_pathname=gui_settings.BASE_HREF,
        external_stylesheets=[
            # dbc.themes.BOOTSTRAP,
            # dbc.icons.BOOTSTRAP,
        ],  # Do not use external_stylesheets
        assets_ignore="styles-dark.css",  # *.css in assets are automatically imported; they need to be explicitly ignored
        use_pages=True,
    )

    # Bootstrap ICONS: https://icons.getbootstrap.com/

    app._favicon = "favicon.ico"
    app.title = "AEGIS visualizer"
    app.layout = layout.get_app_layout()

    return app



from aegis_sim.utilities.container import Container
from aegis_sim.utilities.analysis import survival, reproduction, genome


# x-axis is age
def get_total_survivorship(container: Container, iloc=-1):
    ys = container.get_surv_observed_interval()
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc].cumprod()
    return ys, max_iloc


def get_mortality_observed(container: Container, iloc=-1):
    ys = container.get_surv_observed_interval()
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc].pipe(lambda x: 1 - x)
    return ys, max_iloc


def get_mortality_intrinsic(container: Container, iloc=-1):
    ys = container.get_phenotype_intrinsic_interval("surv")
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc].pipe(lambda x: 1 - x).to_numpy()
    return ys, max_iloc


def get_intrinsic_survivorship(container: Container, iloc=-1):
    ys = container.get_phenotype_intrinsic_interval("surv")
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc].cumprod()
    return ys, max_iloc


def get_fertility(container: Container, iloc=-1):
    ys = container.get_fert_observed_interval()
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc]
    return ys, max_iloc


def get_cumulative_reproduction(container: Container, iloc=-1):
    ys = reproduction.get_cumulative_reproduction(
        container.get_phenotype_intrinsic_interval("repr"),
        container.get_config()["AGE_LIMIT"],
        container.get_config()["MATURATION_AGE"],
    )
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc]
    return ys, max_iloc


def get_birth_table(container: Container, iloc=-1):
    ys = container.get_birth_table_observed_interval()
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc]
    return ys, max_iloc


def get_death_table(container: Container, iloc=-1):
    ys = container.get_death_table_observed_interval().unstack("cause_of_death")
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc].unstack("cause_of_death")
    return ys, max_iloc


def get_death_table_normalized(container: Container, iloc=-1):
    ys, max_iloc = get_death_table(container=container, iloc=iloc)
    ys = ys.div(ys.sum(1), axis=0)
    return ys, max_iloc


def get_life_table(container: Container, iloc=-1):
    ys = container.get_life_table_observed_interval(normalize=True)
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc]
    return ys, max_iloc


# x-axis is step
def get_lifetime_reproduction(container: Container, iloc=None):
    ys = container.get_lifetime_reproduction()
    max_iloc = None
    return ys, max_iloc


def get_life_expectancy(container: Container, iloc=None):
    ys = survival.get_life_expectancy(
        container.get_phenotype_intrinsic_interval("surv"),
        container.get_config()["AGE_LIMIT"],
    )
    max_iloc = None
    return ys, max_iloc


def get_population_size_after_reproduction(container: Container, iloc=None):
    ys = container.get_population_size_after_reproduction().popsize
    max_iloc = None
    return ys, max_iloc


# x-axis is other
def get_derived_allele_freq(container: Container, iloc=-1):
    ys = genome.get_derived_allele_freq(container.get_genotypes_intrinsic_interval())
    max_iloc = ys.shape[0]
    ys = ys.iloc[iloc].to_numpy()
    ys = ys[ys != 0]
    return ys, max_iloc


def get_bit_states(container: Container, iloc=None):
    ys = container.get_genotypes_intrinsic_interval()  # unsorted
    max_iloc = None
    return ys, max_iloc



import setuptools
import pathlib

__version__ = "2.3.0.2"

# Latest pypi version is 2.3.0.2
# Latest test_pypi version is 2.3.0


setuptools.setup(
    name="aegis-sim",
    version=__version__,
    description="Numerical model for life history evolution of age-structured populations",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    keywords=["evolution", "aging", "life history", "agent-based model", "simulation"],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    author="Martin Bagic, Dario Valenzano",
    author_email="martin.bagic@outlook.com, Dario.Valenzano@leibniz-fli.de",
    project_urls={
        "Source": "https://github.com/valenzano-lab/aegis",
        # "Documentation": "",
    },
    url="https://github.com/valenzano-lab/aegis",
    # packages=setuptools.find_packages(include=["aegis", "gui", "documentation", "aegis.*"]),
    packages=setuptools.find_namespace_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "aegis = aegis.__main__:main",
        ]
    },
    install_requires=[
        "numpy",
        "pandas",
        "PyYAML",
        "pyarrow",
        "platformdirs",
        "dash>=2.17.1",  # GUI
        "dash-bootstrap-components",  # GUI
        "urllib3==1.26.19",
        "psutil",
        "kaleido",  # for image export for dash figures TODO use a simpler method
    ],
    extras_require={
        "dev": [
            "pytest==6.2.4",
            "flake8",
            "black",
            "tabulate",
            "pdoc",
        ]
    },
)



import logging



class GPM:
    """Genotype-phenotype map

    Order of elements in the vector does not matter. # TODO Explain better

    ### GENOTYPE-PHENOTYPE MAP (GPM) ###
    In AEGIS, every individual carries a genome which encodes an intrinsic phenotype.
    A genome can be converted into an intrinsic phenotype using the genotype-phenotype map (GPM).
    Conceptually, the GPM contains the information on how each site affects the intrinsic phenotype
    of the individual (e.g. the first site decreases fertility by 0.15% at age class 28).

    The GPM can be saved in two formats: a list or a matrix.

    If it is a list, it will be a list of quadruple (4-tuple) with the following structure: `index`, `trait`, `age`, `magnitude`.
    Thus a single quadruple encodes an effect of a single site at the index `index` (e.g. 1)
    on the trait `trait` (e.g. fertility) expressed at the age `age` (e.g. 28). The change to the trait is of magnitude `magnitude` (0.85).
    When a site is pleiotropic, there will be multiple quadruples with the same `index`.
    We distringuish between age-pleiotropy (a single site affecting at least one trait at multiple ages) and trait-pleiotropy (a single site affecting multiple traits).

    If the GPM is encoded as a matrix, it is a 3D matrix where dimensions encode `index`, `trait` and `age`,
    while the matrix values encode the `magnitude`s.

    When most sites are age-pleiotropic and trait-pleiotropic, the optimal encoding format is a matrix.
    When most sites have age-specific and trait-specific effects, the optimal encoding format is a list
    rather than a matrix because the matrix will be very sparse (it will carry a lot of 0's).
    """

    def __init__(self, AGE_LIMIT, phenomatrix, phenolist):
        self.AGE_LIMIT = AGE_LIMIT
        self.phenomatrix = phenomatrix
        self.phenolist = phenolist

        self.dummy = self.phenolist == [] and self.phenomatrix is None
        if self.dummy:
            logging.info("Phenomap inactive.")

    def phenodiff(self, vectors, zeropheno):
        """
        Vectors .. matrix of shape (individual, ?); items are real numbers, chromosomes and loci are resolved
        Phenomatrix ..
        Phenolist .. list of (bit_index, trait, age, magnitude)
        """

        if self.phenomatrix is not None:
            # TODO BUG resolve phenomatrix
            return vectors.dot(self.phenomatrix)

        elif self.phenolist is not None:
            phenodiff = zeropheno.copy()
            for vec_index, trait, age, magnitude in self.phenolist:
                vec_state = vectors[:, vec_index]
                phenotype_change = vec_state * magnitude
                phenotype_index = parameterization.traits[trait].start + age
                phenodiff[:, phenotype_index] += phenotype_change
            return phenodiff

        else:
            raise Exception("Neither phenomatrix nor phenolist has been provided.")

    def add_initial_values(self, array):
        array = array.copy()

        for traitname, trait in parameterization.traits.items():
            array[:, trait.slice] += trait.initpheno
        return array

    def __call__(self, interpretome, zeropheno):
        if self.dummy:
            return interpretome
        else:
            phenodiff = self.phenodiff(interpretome, zeropheno=zeropheno)
            phenowhole = self.add_initial_values(phenodiff)
            return phenowhole



import logging
import yaml
import types

from aegis_sim.parameterization.default_parameters import (
    get_default_parameters,
    DEFAULT_PARAMETERS,
    get_species_parameters,
)


class ParameterManager:
    def init(self, custom_config_path, custom_input_params):
        self.custom_config_path = custom_config_path
        self.custom_input_params = custom_input_params
        self.final_config = None

        self.parameters = self()

    def __call__(self):
        """
        Getting parameters from three sources:
        1. Default
        2. Configuration file
        3. Function arguments

        When a parameter value is specified multiple times, 3 overwrites 2 which overwrites 1.
        """

        default_parameters = get_default_parameters()
        custom_config_params = self.read_config_file()
        self.validate(custom_config_params)
        for k in default_parameters.keys():
            if k in custom_config_params and default_parameters[k] != custom_config_params[k]:
                logging.debug(
                    f"-- {k} is different in config ({custom_config_params[k]}) vs default ({default_parameters[k]})"
                )

        SPECIES_PRESET = custom_config_params.get("SPECIES_PRESET", default_parameters["SPECIES_PRESET"])
        species_config_params = get_species_parameters(SPECIES_PRESET)

        logging.info(f"Using {SPECIES_PRESET} as species preset: " + repr(species_config_params) + ".")

        # Fuse
        params = {}
        params.update(default_parameters)
        params.update(species_config_params)
        params.update(custom_config_params)
        params.update(self.custom_input_params)

        self.final_config = params.copy()

        # convert to types.SimpleNamespace
        params = types.SimpleNamespace(**params)
        logging.info("Final parameters to use in the simulation: " + repr(params) + ".")
        return params

    def read_config_file(self):

        # No configuration file specified
        if self.custom_config_path == "":
            logging.info("No configuration file has been specified.")
            return {}

        # Configuration file specified...
        with open(self.custom_config_path, "r") as file_:
            ccp = yaml.safe_load(file_)

        # ... but it is empty
        if ccp is None:
            logging.info("Configuration file is empty.")
            ccp = {}

        return ccp

    @staticmethod
    def validate(pdict, validate_serverrange=False):
        for key, val in pdict.items():
            # Validate key
            if all(key != p.key for p in DEFAULT_PARAMETERS.values()):
                raise ValueError(f"'{key}' is not a valid parameter name")

            # Validate value type and range
            DEFAULT_PARAMETERS[key].validate_dtype(val)
            DEFAULT_PARAMETERS[key].validate_inrange(val)

            if validate_serverrange:
                DEFAULT_PARAMETERS[key].validate_serverrange(val)





import yaml
import pathlib
from typing import Optional


class GuiSettings:
    ENVIRONMENT: str
    DEBUG_MODE: bool
    # loglevel: str
    SIMULATION_NUMBER_LIMIT: Optional[int]
    # can_delete_default_sim: bool
    # default_selection_states: tuple
    DATA_RETENTION_DAYS: Optional[int]
    ABS_PATH_TO_BASE_DIRECTORY: Optional[str]
    PORT: Optional[int]
    BASE_HREF = "/aegis/"

    def can_run_more_simulations(self, currently_running):
        if self.SIMULATION_NUMBER_LIMIT is None:
            return True
        else:
            return self.SIMULATION_NUMBER_LIMIT > currently_running

    def get_base_dir(self):
        if self.ABS_PATH_TO_BASE_DIRECTORY is None:
            return pathlib.Path.home().absolute()
        else:
            return pathlib.Path(self.ABS_PATH_TO_BASE_DIRECTORY).absolute()

    def set(self, environment: str, debug):

        # Read from YML
        path_to_yaml = pathlib.Path(__file__).parent / "gui_settings.yml"
        with open(path_to_yaml, "r") as file:
            yml = yaml.safe_load(file)
        if environment == "local":
            yml = yml["LocalConfig"]
        elif environment == "server":
            yml = yml["ServerConfig"]
        else:
            raise ValueError(f"{environment} is an invalid input for environment; should be 'local' or 'server'")

        assert isinstance(debug, bool)

        self.ENVIRONMENT = yml["ENVIRONMENT"]
        self.DEBUG_MODE = debug
        self.SIMULATION_NUMBER_LIMIT = yml["SIMULATION_NUMBER_LIMIT"]
        self.DATA_RETENTION_DAYS = yml["DATA_RETENTION_DAYS"]
        self.MAX_SIM_NAME_LENGTH = yml["MAX_SIM_NAME_LENGTH"]
        self.ABS_PATH_TO_BASE_DIRECTORY = yml["ABS_PATH_TO_BASE_DIRECTORY"]
        self.PORT = yml["PORT"]

        # Derived
        self.base_dir = self.get_base_dir() / "aegis_data"
        self.sim_dir = self.base_dir / "sim_data"
        self.figure_dir = self.base_dir / "figures"

        self.base_dir.mkdir(exist_ok=True, parents=True)
        self.sim_dir.mkdir(exist_ok=True, parents=True)
        self.figure_dir.mkdir(exist_ok=True, parents=True)

    def wrap_href(self, href):
        return self.BASE_HREF + href


gui_settings = GuiSettings()

# def create_config_class(name, attrs):
#     """Dynamically create a configuration class."""
#     return type(name, (BaseConfig,), attrs)


# def load_configs_from_yaml(file_path):

#     config_classes = {}

#     for config_name, config_attrs in configs.items():
#         config_class = create_config_class(config_name, config_attrs)
#         config_classes[config_name] = config_class

#     return config_classes


# LocalConfig = config_classes["LocalConfig"]
# ServerConfig = config_classes["ServerConfig"]
